import math
import multiprocessing
from typing import List, Callable, Dict, Any
from pydantic import BaseModel
from icolos.utils.enums.parallelization import ParallelizationEnum

_PE = ParallelizationEnum


class Subtask(BaseModel):
    status: _PE = _PE.STATUS_READY
    times_tried: int = 0
    data: Any
    job_id: str = None

    def increment_tries(self):
        self.times_tried += 1

    def set_status(self, status: str):
        self.status = status

    def set_job_id(self, job_id: str):
        self.job_id = job_id

    def set_status_failed(self):
        self.set_status(_PE.STATUS_FAILED)

    def set_status_success(self):
        self.set_status(_PE.STATUS_SUCCESS)


class SubtaskContainer(BaseModel):
    max_tries: int
    subtasks: List[Subtask] = []

    def __init__(self, **data):
        super().__init__(**data)

    def clear(self):
        self.subtasks = []

    def load_data(self, data: List[Any]):
        self.clear()
        self.add_data(data=data)

    def add_data(self, data: List[Any]):
        for data_element in data:
            self.subtasks.append(
                Subtask(status=_PE.STATUS_READY, times_tried=0, data=data_element)
            )

    def get_todo_tasks(self) -> List[Subtask]:
        todo_subtasks = []
        for subtask in self.subtasks:
            if (
                subtask.status == _PE.STATUS_READY
                or subtask.status == _PE.STATUS_FAILED
            ) and subtask.times_tried < self.max_tries:
                todo_subtasks.append(subtask)
        return todo_subtasks

    def get_done_tasks(self) -> List[Subtask]:
        done_subtasks = []
        for subtask in self.subtasks:
            if subtask.status == _PE.STATUS_SUCCESS or (
                subtask.times_tried >= self.max_tries
                and subtask.status not in (_PE.STATUS_RUNNING, _PE.STATUS_READY)
            ):
                done_subtasks.append(subtask)
        return done_subtasks

    def get_running_tasks(self) -> List[Subtask]:
        running_subtasks = []
        for subtask in self.subtasks:
            if subtask.status == _PE.STATUS_RUNNING:
                running_subtasks.append(subtask)
        return running_subtasks

    def get_sublists(
        self, partitions=None, slice_size=None, get_first_n_lists=None
    ) -> List[List[Subtask]]:
        if partitions is None and slice_size is None:
            raise ValueError("Either specify partitions or slice size.")

        # only get tasks that are not yet completed or have some tries left
        subtasks = self.get_todo_tasks()

        # decide on the chunk size, either by doing partitions or by specifying the slice size directly
        sublists = []
        if partitions is not None:
            chunk_size = int(math.ceil(len(subtasks) / partitions))
        else:
            chunk_size = slice_size

        # wrap the tasks in lists as required
        for i in range(0, len(subtasks), chunk_size):
            sublist = []
            for j in range(i, min(i + chunk_size, len(subtasks))):
                sublist.append(subtasks[j])
            sublists.append(sublist)

        if get_first_n_lists is not None and len(sublists) > get_first_n_lists:
            return sublists[:get_first_n_lists]
        else:
            return sublists

    def done(self) -> bool:
        for subtask in self.subtasks:
            if subtask.status == _PE.STATUS_SUCCESS:
                continue
            if subtask.status == _PE.STATUS_READY or (
                subtask.status == _PE.STATUS_FAILED
                and subtask.times_tried < self.max_tries
            ):
                return False
        return True

    def any_failed(self) -> bool:
        if len(
            [True for subtask in self.subtasks if subtask.status == _PE.STATUS_FAILED]
        ):
            return True
        return False

    def set_max_tries(self, max_tries: int):
        self.max_tries = max_tries

    def __len__(self) -> int:
        return len(self.subtasks)


class Parallelizer(BaseModel):
    func: Callable

    def __init__(self, **data):
        super().__init__(**data)

    def rearrange_input(self, inp_dict: Dict[str, List]) -> List[Dict]:
        return [dict(zip(inp_dict, ele)) for ele in zip(*inp_dict.values())]

    def execute_parallel(self, **kwargs):
        # translate the dictionary with the lists of arguments into a list of individual dictionaries
        # e.g. {'one': [1, 2, 3], 'two': ['aaaa', 'bbb', 'cc'], 'three': [0.2, 0.2, 0.1]} --->
        # [{'one': 1, 'two': 'aaaa', 'three': 0.2},
        #  {'one': 2, 'two': 'bbb', 'three': 0.2},
        #  {'one': 3, 'two': 'cc', 'three': 0.1}]
        list_exec = self.rearrange_input(kwargs)

        # # run in parallel; wait for all subjobs to finish before proceeding
        processes = []
        for subprocess_args in list_exec:
            p = multiprocessing.Process(target=self.func, kwargs=subprocess_args)
            processes.append(p)
            p.start()

        for p in processes:
            p.join()



from pydantic import BaseModel
from icolos.core.workflow_steps.step import StepBase
from icolos.utils.enums.step_enums import StepBaseEnum

_SBE = StepBaseEnum


class StepIOBase(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)



import os
import sys
from icolos.utils.enums.program_parameters import OpenBabelEnum
from icolos.utils.execute_external.execute import ExecutorBase

EE = OpenBabelEnum()


class OpenBabelExecutor(ExecutorBase):
    """For the execution of the "obabel" binary."""

    def __init__(self):
        # in case the environment is not activated, add the path to the binary here
        obabel_location = os.path.dirname(sys.executable)
        super().__init__(prefix_execution=None, binary_location=obabel_location)

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        # check, whether a proper executable is provided
        if command not in [EE.OBABEL]:
            raise ValueError(
                "Parameter command must be an element of the internal OpenBabel executable list."
            )

        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=location,
            pipe_input=pipe_input,
        )

    def is_available(self):
        # unfortunately, "obabel" does not return a meaningful return value (always '1'), so instead try to parse
        # the "stdout" of the standard message; note, that "OpenBabel" is part of the environment and should always work
        try:
            result = self.execute(command=EE.OBABEL, arguments=[], check=False)
            if EE.OBABEL_IDENTIFICATION_STRING in result.stdout:
                return True
            return False
        except Exception as e:
            return False



import json
import logging.config as logging_config
from icolos.loggers.entrypoint_logger import EntryPointLogger
from icolos.utils.enums.composite_agents_enums import WorkflowEnum
from icolos.utils.enums.logging_enums import LoggingConfigEnum

from icolos.utils.general.convenience_functions import *

_WE = WorkflowEnum()
_LE = LoggingConfigEnum()


def initialize_logging(log_conf_path: str, workflow_conf: dict) -> EntryPointLogger:
    with open(log_conf_path, "r") as f:
        log_conf_dict = json.load(f)
    header = nested_get(workflow_conf, [_WE.WORKFLOW, _WE.HEADER], default={})
    if in_keys(header, [_WE.LOGGING, _WE.LOGGING_LOGFILE]):
        try:
            log_conf_dict["handlers"]["file_handler"]["filename"] = nested_get(
                header, [_WE.LOGGING, _WE.LOGGING_LOGFILE], None
            )
            log_conf_dict["handlers"]["file_handler_blank"]["filename"] = nested_get(
                header, [_WE.LOGGING, _WE.LOGGING_LOGFILE], None
            )
        except KeyError:
            pass
    logging_config.dictConfig(log_conf_dict)
    logger = EntryPointLogger()
    return logger



from subprocess import CompletedProcess
import time

from icolos.core.containers.generic import GenericContainer, GenericData
import multiprocessing
import shutil
import tempfile
from typing import Callable, List, Dict, Optional, Tuple

from pydantic import BaseModel, PrivateAttr
from rdkit import Chem
from copy import deepcopy
import os
from icolos.core.containers.gmx_state import GromacsState
from icolos.core.containers.perturbation_map import PerturbationMap


from icolos.core.step_utils.input_preparator import (
    StepData,
    InputPreparator,
    StepInputParameters,
)
from icolos.loggers.steplogger import StepLogger
from icolos.loggers.blank_logger import BlankLogger
from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.core.containers.compound import Compound, Conformer
from icolos.core.step_utils.step_writeout import (
    StepWriteoutParameters,
    WriteOutHandler,
    _SBE,
)
from icolos.utils.enums.execution_enums import (
    ExecutionPlatformEnum,
)
from icolos.utils.execute_external.execute import Executor
from icolos.utils.execute_external.slurm_executor import SlurmExecutor
from icolos.utils.general.icolos_exceptions import StepFailed

from icolos.utils.enums.compound_enums import CompoundTagsEnum
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.write_out_enums import WriteOutEnum
from icolos.utils.general.files_paths import gen_tmp_file, any_in_file
from icolos.utils.general.parallelization import SubtaskContainer, Subtask
from tempfile import mkdtemp
from distutils.dir_util import copy_tree
from icolos.core.containers.compound import unroll_enumerations, unroll_conformers
from icolos.utils.general.progress_bar import get_progress_bar_string

_LE = LoggingConfigEnum()
_WE = WriteOutEnum()
_CTE = CompoundTagsEnum()
_SGE = StepGromacsEnum()
_EPE = ExecutionPlatformEnum


class StepFailurePolicyParameters(BaseModel):
    n_tries: int = 1
    retry_wait_seconds: int = 10


class StepExecutionResourceParameters(BaseModel):
    partition: str = _EPE.CORE
    time: str = "12:00:00"
    gres: str = None
    tasks: str = None
    mem: str = None
    cores: int = None
    modules: List = []
    other_args: dict = {}
    additional_lines: List = []


class StepExecutionParameters(BaseModel):
    class StepExecutionParallelizationParameters(BaseModel):
        jobs: int = 1
        max_length_sublists: int = None

    prefix_execution: str = None
    binary_location: str = None
    pipe_input: str = None
    parallelization: StepExecutionParallelizationParameters = (
        StepExecutionParallelizationParameters()
    )
    failure_policy: StepFailurePolicyParameters = StepFailurePolicyParameters()
    check_backend_availability: bool = False
    resources: StepExecutionResourceParameters = StepExecutionResourceParameters()
    platform: _EPE = _EPE.LOCAL


class StepSettingsArgsParameters(BaseModel):
    flags: List = []
    parameters: Dict = {}


class StepSettingsParameters(BaseModel):
    arguments: StepSettingsArgsParameters = StepSettingsArgsParameters()
    additional: Dict = {}


class StepBase(BaseModel):
    step_id: str
    work_dir: str = None
    type: str = None
    data: StepData = StepData()
    input: StepInputParameters = StepInputParameters()
    writeout: List[StepWriteoutParameters] = []
    execution: StepExecutionParameters = StepExecutionParameters()
    settings: StepSettingsParameters = StepSettingsParameters()

    class Config:
        underscore_attrs_are_private = True

    _logger = PrivateAttr()
    _logger_blank = PrivateAttr()
    _old_wdir = PrivateAttr()
    _workflow_object = PrivateAttr()
    _backend_executor: Executor = PrivateAttr()
    _subtask_container: SubtaskContainer = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)

        self._logger_blank = BlankLogger()
        self._old_wdir = os.getcwd()
        self._workflow_object = None
        self._backend_executor = None

        self._logger = StepLogger()
        self._logger_blank = BlankLogger()

    # @staticmethod
    def _make_tmpdir(self):
        if self.work_dir is not None:
            self._logger.log(f"Using specified work_dir {self.work_dir}", _LE.DEBUG)
            return self.work_dir
        else:
            self.work_dir = tempfile.mkdtemp()
            return self.work_dir

    def _remove_temporary(self, paths):
        if paths is not None:
            if not isinstance(paths, list):
                paths = [paths]
            if (
                self.get_workflow_object() is None
                or self.get_workflow_object().header.global_settings.remove_temporary_files
            ):
                for path in paths:
                    if os.path.isdir(path):
                        shutil.rmtree(path, ignore_errors=True)
                    elif os.path.isfile(path) and os.path.exists(path):
                        os.remove(path)
                    else:
                        self._logger.log(
                            f"Path {path} is neither a valid folder nor file path.",
                            _LE.WARNING,
                        )
            else:
                self._logger.log(
                    f"Keeping {len(paths)} temporary file(s) / folder(s): {', '.join(paths)}",
                    _LE.DEBUG,
                )

    @staticmethod
    def _move_to_temp_dir() -> str:
        cur_tmp_dir = tempfile.mkdtemp()
        os.chdir(cur_tmp_dir)
        return cur_tmp_dir

    @staticmethod
    def _move_to_dir(path: str):
        os.chdir(path)

    def _restore_working_dir(self):
        os.chdir(self._old_wdir)

    def execute(self):
        raise NotImplementedError

    def get_compound_by_name(self, name: str) -> Optional[Compound]:
        for compound in self.data.compounds:
            if compound.get_name() == name:
                return compound

    def get_compounds(self) -> List[Compound]:
        return self.data.compounds

    def get_generic(self) -> GenericContainer:
        return self.data.generic

    def clone_compounds(self) -> List[Compound]:
        return [deepcopy(comp) for comp in self.data.compounds]

    def process_write_out(self):
        for writeout in self.writeout:
            writeout_handler = WriteOutHandler(config=writeout)
            writeout_handler.set_data(self.data)
            # attach workflow data at this point
            writeout_handler.set_workflow_data(self.get_workflow_object().workflow_data)
            writeout_handler.write()

    def get_compound_stats(self) -> Tuple[int, int, int]:
        n_comp = len(self.get_compounds())
        n_enum = len(unroll_enumerations(self.get_compounds()))
        n_conf = len(unroll_conformers(self.get_compounds()))
        return n_comp, n_enum, n_conf

    def generate_input(self):
        preparator = InputPreparator(
            workflow=self.get_workflow_object(), logger=self._logger
        )
        self.data, self.work_dir = preparator.generate_input(
            step_input=self.input, step_type=self.type
        )

        # check for a perturbation map for fep workflows
        self._logger.log(
            f"Loaded {len(self.data.compounds)} compounds and {len(self.data.generic.get_flattened_files())} generic data fields for step {self.get_step_id()}.",
            _LE.DEBUG,
        )

    def set_workflow_object(self, workflow_object):
        self._workflow_object = workflow_object

    def get_workflow_object(self):
        return self._workflow_object

    def get_perturbation_map(self) -> PerturbationMap:
        return self._workflow_object.workflow_data.perturbation_map

    def get_step_id(self) -> str:
        return self.step_id

    def set_step_id(self, step_id: str):
        self.step_id = step_id

    def _initialize_backend(self, executor: Callable):
        if self.execution.platform == _EPE.SLURM:
            self._backend_executor = SlurmExecutor(
                prefix_execution=self.execution.prefix_execution,
                binary_location=self.execution.binary_location,
                cores=self.execution.resources.cores,
                tasks=self.execution.resources.tasks,
                partition=self.execution.resources.partition,
                time=self.execution.resources.time,
                mem=self.execution.resources.mem,
                modules=self.execution.resources.modules,
                other_args=self.execution.resources.other_args,
                additional_lines=self.execution.resources.additional_lines,
                gres=self.execution.resources.gres,
            )
        else:
            self._backend_executor = executor(
                prefix_execution=self.execution.prefix_execution,
                binary_location=self.execution.binary_location,
            )

    def _unroll_compounds(
        self,
        compounds: List[Compound],
        level: str = _SBE.WRITEOUT_COMP_CATEGORY_CONFORMERS,
    ) -> List[Conformer]:
        # TODO: move this to step_base or merge with methods from compound itself

        all_conformers = []
        for comp in compounds:
            for enum in comp.get_enumerations():
                if level == _SBE.WRITEOUT_COMP_CATEGORY_ENUMERATIONS:
                    all_conformers.append(enum)
                elif level == _SBE.WRITEOUT_COMP_CATEGORY_CONFORMERS:
                    for conf in enum:
                        all_conformers.append(conf)
        return all_conformers

    def write_conformers(self, path: str):
        """Convenience function for frequent conformer coordinate write-out. Better to use the WriteOutHandler class."""
        compounds_copy = self.clone_compounds()
        params = {
            _SBE.WRITEOUT_CONFIG: {
                _SBE.WRITEOUT_COMP: {
                    _SBE.WRITEOUT_COMP_CATEGORY: _SBE.WRITEOUT_COMP_CATEGORY_CONFORMERS
                },
                _SBE.WRITEOUT_DESTINATION: {
                    _SBE.WRITEOUT_DESTINATION_RESOURCE: path,
                    _SBE.WRITEOUT_DESTINATION_TYPE: _SBE.WRITEOUT_DESTINATION_TYPE_FILE,
                    _SBE.WRITEOUT_DESTINATION_FORMAT: _SBE.FORMAT_SDF,
                },
            }
        }
        writeout_handler = WriteOutHandler(**params)
        writeout_handler.set_data(StepData(compounds=compounds_copy))
        writeout_handler.write()

    def write_generic_by_extension(self, path: str, ext: str, join=True):
        """Write all generic data objects with a given extension

        :param str path: path to write object to, by default just a directory
        :param str ext: extension
        :param bool join: controls join behaviour, if True, joins existing filename to path, defaults to True
        """
        for file in self.data.generic.get_files_by_extension(ext):
            file.write(path, join=join)

    def write_generic_by_name(self, path: str, name: str):
        """Write a generic file by name

        :param str path: directory to write to
        :param str name: name of file to be written out
        """
        file = self.data.generic.get_file_by_name(name)
        file.write(path)

    def _check_backend_availability(self, strict=True):
        if self._backend_executor is None:
            raise Exception(
                "Cannot check backend availability before initialization is complete."
            )

        if self.execution.check_backend_availability:
            if not self._backend_executor.is_available():
                if strict:
                    raise StepFailed(
                        f"Cannot initialize backend for step {self.step_id} - abort."
                    )
                else:
                    self._logger.log(
                        f"Backend availability check failed, proceeding anyways.",
                        _LE.WARNING,
                    )
            else:
                self._logger.log(f"Checked backend availability - valid.", _LE.DEBUG)

    def _input_object_valid(self, obj) -> bool:
        if obj.get_molecule() is None or not isinstance(obj.get_molecule(), Chem.Mol):
            self._logger.log(
                f"Object {obj.get_index_string()} skipped - no valid molecule.",
                _LE.WARNING,
            )
            return False
        return True

    def _input_object_empty(self, obj) -> bool:
        if obj.empty():
            self._logger.log(
                f"Object {obj.get_index_string()} is skipped (empty).", _LE.WARNING
            )
            return True
        return False

    # TODO: REMOVE THIS FUNCTION (see: write_molecule_to_sdf())
    def _prepare_temp_input(self, tmp_dir: str, molecule: Chem.Mol) -> str:
        _, tmp_sdf_path = gen_tmp_file(suffix=".sdf", dir=tmp_dir)
        if molecule is None or not isinstance(molecule, Chem.Mol):
            raise ValueError(
                "Function requires input attribute to be an RDkit molecule."
            )
        writer = Chem.SDWriter(tmp_sdf_path)
        writer.write(molecule)
        writer.close()
        self._logger.log(f"Wrote input molecule to file {tmp_sdf_path}.", _LE.DEBUG)
        return tmp_sdf_path

    def _get_sublists(self, get_first_n_lists: int = None) -> List[List[Subtask]]:
        number_cores = self._get_number_cores()

        # decide how to slice the ligand list depending on whether a maximum length is defined or not
        if self.execution.parallelization.max_length_sublists is not None:
            slice_size = min(
                max(self.execution.parallelization.max_length_sublists, 1),
                len(self._subtask_container),
            )
            return self._subtask_container.get_sublists(
                partitions=None,
                slice_size=slice_size,
                get_first_n_lists=get_first_n_lists,
            )
        else:
            # split the ligands into as many cores as available
            partitions = min(number_cores, len(self._subtask_container))
            return self._subtask_container.get_sublists(
                partitions=partitions,
                slice_size=None,
                get_first_n_lists=get_first_n_lists,
            )

    def _get_number_cores(self):
        # prepare the parallelization and set the number of cores to be used
        cores = self.execution.parallelization.jobs
        if cores == 0:
            cores = 1
        elif cores < 0:
            # subtract the number of cores (neg. value, thus add up) from total number of cores, e.g. -1 will
            # use all available cores minus 1
            cores = multiprocessing.cpu_count() + cores
        if cores > multiprocessing.cpu_count():
            self._logger.log(
                f"WARNING: running {cores} processes on {multiprocessing.cpu_count()} logical cores!",
                _LE.WARNING,
            )
        return cores
    
    def get_arguments(self, defaults: dict = None) -> list:
        """
        Construct pmx-specific arguments from the step defaults,
        overridden by arguments specified in the config file
        """
        arguments = []

        # add flags
        for flag in self.settings.arguments.flags:
            arguments.append(flag)

        # flatten the dictionary into a list for command-line execution
        for key in self.settings.arguments.parameters.keys():
            arguments.append(key)
            arguments.append(self.settings.arguments.parameters[key])

        # add defaults, if not already present
        if defaults is not None:
            for key, value in defaults.items():
                if key not in arguments:
                    arguments.append(key)
                    arguments.append(value)
        return arguments

    def _print_log_file(self, path: str):
        if os.path.isfile(path):
            with open(path, "r") as log_file:
                log_file_raw = log_file.readlines()
                self._logger.log(f"Printing log file {path}:\n", _LE.DEBUG)
                for line in log_file_raw:
                    self._logger_blank.log(line.rstrip("\n"), _LE.DEBUG)
                self._logger_blank.log("", _LE.DEBUG)
                self._logger.log("--- End file", _LE.DEBUG)

    def _add_data_to_generic(self, file, data, extension=None):
        """Write data from arbitrary file to generic container class"""
        file_name = file.split("/")[-1]
        # file types where they can be passed as arguments in a subsequent step
        # TODO: this is not maintainable!
        file_tag = (
            True
            if file.endswith((".gro", "topol.top", "tpr", "fmp", "edge"))
            else False
        )
        file = GenericData(
            file_name=file_name, file_data=data, argument=file_tag, extension=extension
        )
        self.data.generic.add_file(file)

    def _parse_output(
        self,
        tmp_dir,
        exclusion_list=(
            "#",
            "AC",
            "AC0",
            "INF",
            "hashed",
            "metadata",
            "timekeys",
            "000000",
        ),
    ):
        """Generic method for parsing generic writeout, can be overwritten in child classes"""
        self.data.generic.clear_file_dict()
        file_list = [os.path.join(tmp_dir, f) for f in os.listdir(tmp_dir)]
        for file in file_list:
            if os.path.isfile(file) and not file.endswith(exclusion_list):
                try:
                    with open(file, "r") as f:
                        data = f.read()
                    binary = False
                except UnicodeDecodeError:
                    with open(file, "rb") as f:
                        data = f.read()
                    binary = True
                # work out if we handle the data or just the path to it on disk
                file_size = os.stat(file).st_size
                if file_size > float(_SBE.FILE_SIZE_THRESHOLD.value):
                    # do not write to the dict - file is too large to store in memory
                    _, tmp_path = gen_tmp_file(suffix="." + str(file).split(".")[-1])
                    self._logger.log(
                        f"Large file detected, storing at {tmp_path}", _LE.INFO
                    )
                    if binary:
                        with open(tmp_path, "wb") as f:
                            f.write(data)
                    else:
                        with open(tmp_path, "w") as f:
                            f.write(data)
                    data = tmp_path

                self._add_data_to_generic(file, data)
                self._logger.log(f"Stored data for file {file}", _LE.DEBUG)
            elif os.path.isdir(file):
                tmp_dir = mkdtemp()
                copy_tree(file, tmp_dir)
                self._add_data_to_generic(file=file, data=tmp_dir, extension="dir")

                # we have picked up a directory, we want the entire contents copied somewhere

    def _wait_until_file_generation(
        self,
        path,
        path_log=None,
        interval_sec=1,
        maximum_sec=None,
        success_strings: set = set(),
        fail_strings: set = set(),
    ) -> bool:
        # TODO: Refactor that without breaking the Glide dependency.
        counter = 0
        while not os.path.exists(path):
            # wait for an interval
            time.sleep(interval_sec)
            counter = counter + 1

            # if a Glide logfile path has been specified, check, whether critical messages indicating an abort are there
            # note, that we return "True" to indicate that the "file generation" has nevertheless been completed
            if path_log is not None:
                if any_in_file(path_log, fail_strings):
                    self._logger.log(
                        f"A critical error occurred in sublist execution.", _LE.WARNING
                    )
                    self._print_log_file(path_log)
                    return True
                if any_in_file(path_log, success_strings):
                    # log file indicates job is done; give a bit of leeway to ensure the writing is done
                    time.sleep(3)
                    break

            # if there's time left, proceed
            if maximum_sec is not None and counter * interval_sec >= maximum_sec:
                break
        if os.path.exists(path):
            return True
        else:
            return False

    def _log_execution_progress(self):
        number_tasks_done = len(self._subtask_container.get_done_tasks())
        number_tasks_total = len(self._subtask_container.subtasks)
        self._logger.log(
            get_progress_bar_string(number_tasks_done, number_tasks_total, length=65),
            _LE.INFO,
        )

    def _log_result(self, result: CompletedProcess):
        """
        logs stdout from completed process to file
        """
        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)

    def _get_additional_setting(self, key: str, default: str = None):
        """
        Query settings.additional with the key, if not set use the default
        """
        return (
            self.settings.additional[key]
            if key in self.settings.additional.keys()
            else default
        )

    def get_topol(self) -> GromacsState:
        if not self.data.generic.get_file_names_by_extension("pkl"):
            return self.data.gmx_state
        else:
            return self.load_topol(
                self.data.generic.get_argument_by_extension("pkl", rtn_file_object=True)
            )



from icolos.core.workflow_steps.step import StepBase
from icolos.utils.enums.step_enums import StepBaseEnum
from icolos.utils.general.convenience_functions import nested_get
from icolos.utils.enums.step_initialization_enum import StepInitializationEnum
from icolos.utils.enums.flow_control_enums import FlowControlInitializationEnum

_IE = StepInitializationEnum()
_FCE = FlowControlInitializationEnum()


def initialize_step_from_dict(step_conf: dict) -> StepBase:
    _STE = StepBaseEnum
    step_type = nested_get(step_conf, _STE.STEP_TYPE, default=None)
    step_type = None if step_type is None else step_type.upper()
    if step_type in _IE.STEP_INIT_DICT.keys():
        return _IE.STEP_INIT_DICT[step_type](**step_conf)
    elif step_type in _FCE.FLOW_CONTROL_INIT_DICT.keys():
        return _FCE.FLOW_CONTROL_INIT_DICT[step_type](**step_conf)
    else:
        raise ValueError(
            f"Backend for step {nested_get(step_conf, _STE.STEPID, '')} unknown."
        )



from pydantic import BaseModel
from icolos.core.workflow_steps.step import StepBase
from Bio import SeqIO
from Bio.PDB import PDBIO
import os
import PeptideBuilder
from PeptideBuilder import Geometry


class StepPeptideEmbedder(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    def execute(self):
        # use the PeptideBuilder python library to build a rough peptide structure using
        # sensible psi, phi angles etc, for subsequent simulation

        tmp_dir = self._make_tmpdir()
        self.data.generic.write_out_all_files(tmp_dir)
        # Extract the peptide sequence from the provided fasta file
        fasta_file = self.data.generic.get_argument_by_extension("fasta")

        sequences = list(SeqIO.parse(os.path.join(tmp_dir, fasta_file), format="fasta"))

        for idx, seq in enumerate(sequences):
            geom = [Geometry.geometry(aa) for aa in seq.seq]
            structure = PeptideBuilder.make_structure_from_geos(geom)

            out = PDBIO()
            out.set_structure(structure)
            # TODO: find a better naming strategy than this
            out.save(os.path.join(tmp_dir, f"sequence_{idx}.pdb"))

        self._parse_output(tmp_dir)

        self._remove_temporary(tmp_dir)



from enum import Enum


class ParallelizationEnum(str, Enum):

    STATUS_READY = "ready"
    STATUS_RUNNING = "running"
    STATUS_SUCCESS = "success"
    STATUS_FAILED = "failed"

    # try to find the internal value and return
    # def __getattr__(self, name):
    #     if name in self:
    #         return name
    #     raise AttributeError

    # # prohibit any attempt to set any values
    # def __setattr__(self, key, value):
    #     raise ValueError("No changes allowed.")



from icolos.core.containers.generic import GenericData
from icolos.core.containers.gmx_state import GromacsState
from icolos.utils.enums.step_enums import StepGromacsEnum
from pydantic import BaseModel
import os
from typing import List
from icolos.core.workflow_steps.step import StepBase
from icolos.core.workflow_steps.step import _LE
import re
from copy import deepcopy
from distutils.dir_util import copy_tree
from icolos.utils.enums.program_parameters import GromacsEnum
import dill

_SGE = StepGromacsEnum()
_GE = GromacsEnum()


class StepGromacsBase(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    def write_input_files(self, tmp_dir: str, topol: GromacsState = None):
        """
        Defaults to writing all available data from the workflow's topology object, unless that file has been specified in generic data, which will be prioritised
        """

        # Normally this should be handled by setting GMXLIB env variable, but for some programs (gmx_MMPBSA), this doesn't work and non-standard forcefields
        # need to be in the working directory
        if _SGE.FORCEFIELD in self.settings.additional:
            copy_tree(
                self.settings.additional[_SGE.FORCEFIELD],
                os.path.join(
                    tmp_dir, self.settings.additional[_SGE.FORCEFIELD].split("/")[-1]
                ),
            )
            self._logger.log(
                f"Copied forcefield at {self.settings.additional[_SGE.FORCEFIELD]} to the working "
                f"directory at {tmp_dir}",
                _LE.INFO,
            )

        self._logger.log(
            f"Writing input files to working directory at {tmp_dir}", _LE.DEBUG
        )
        # if we have specified a topol object, write out these files unless overwritten by an argument from the config
        if topol is not None:
            # handle structure files
            if not self.data.generic.get_files_by_extension("gro"):
                if topol.structures:
                    topol.write_structure(tmp_dir)
            if not self.data.generic.get_files_by_extension("top"):
                if topol.top_lines:
                    topol.write_topol(tmp_dir)
            if not self.data.generic.get_files_by_extension("tpr"):
                if topol.tprs:
                    topol.write_tpr(tmp_dir)
            if not self.data.generic.get_files_by_extension("xtc"):
                if topol.trajectories:
                    topol.write_trajectory(tmp_dir)
            if not self.data.generic.get_files_by_extension("ndx"):
                if topol.ndx:
                    topol.write_ndx(tmp_dir)

        for file in self.data.generic.get_flattened_files():
            file.write(tmp_dir)

    def pickle_topol(self, topol: GromacsState, tmp_dir: str):
        """
        Write the Icolos internal state to a pickled file in the tmpdir
        """
        with open(os.path.join(tmp_dir, "icolos_topol_state.pkl"), "wb") as f:
            dill.dump(topol, f)

    def load_topol(self, file: GenericData) -> GromacsState:
        data = file.get_data()
        return dill.load(data)

    def _parse_arguments(self, flag_dict: dict, args: list = None) -> List:
        arguments = args if args is not None else []
        # first add the settings from the command line
        for key in self.settings.arguments.parameters.keys():
            arguments.append(key)
            arguments.append(str(self.settings.arguments.parameters[key]))
        for flag in self.settings.arguments.flags:
            arguments.append(str(flag))
        for key, value in flag_dict.items():
            # only add defaults if they have not been specified in the json
            if key not in arguments:
                arguments.append(key)
                arguments.append(value)
        return arguments

    def _copy_fields_dict(self):
        try:
            update_dictionary = deepcopy(self.settings.additional[_SGE.FIELDS])
            return update_dictionary
        except KeyError:
            self._logger.log(
                "Update dictionary not present, will use provided mdp file without further modification",
                _LE.WARNING,
            )
            return {}

    def generate_output_file(self, in_file):
        parts = in_file.split(".")
        return parts[0] + "_out" + "." + parts[1]

    def _modify_config_file(
        self, tmp_dir: str, config_file: GenericData, update_dict: dict
    ):
        file_data = config_file.get_data()
        for key, value in update_dict.items():
            pattern = rf"({key})(\s*=\s*)[a-zA-Z0-9\s\_]*(\s*;)"
            pattern = re.compile(pattern)
            matches = re.findall(pattern, file_data)
            if len(matches) == 0:
                self._logger.log(
                    f"Specified key {key} was not found in the mdp file, value was not changed!",
                    _LE.WARNING,
                )
            else:

                file_data = re.sub(pattern, rf"\1\2 {value} \3", file_data)
                self._logger.log(
                    f"Replaced field {key} of mdp file with value {value}", _LE.DEBUG
                )
        self._logger.log(f"Final MDP file for step {self.step_id} is: ", _LE.DEBUG)
        for line in file_data.split("\n"):
            self._logger_blank.log(line, _LE.DEBUG)
        config_file.set_data(file_data)
        config_file.write(tmp_dir)

    def _generate_index_groups(self, tmp_dir: str):
        # dump the first structure file to the tmpdir
        structure = self.get_topol().structures[0]
        structure.write(tmp_dir)
        args = ["-f", structure.get_file_name()]
        ndx_list = [f for f in os.listdir(tmp_dir) if f.endswith(_SGE.FIELD_KEY_NDX)]
        if len(ndx_list) == 1:
            args.extend(["-n", ndx_list[0]])
        result = self._backend_executor.execute(
            command=_GE.MAKE_NDX,
            arguments=args,
            location=tmp_dir,
            check=True,
            pipe_input='echo -e "q"',
        )
        self.get_topol().set_ndx(tmp_dir)
        return result

    def construct_pipe_arguments(self, tmp_dir, params) -> str:
        """
        Constructs the pipe arguments to be passed to gromacs interactive programs
        """
        # look up the groups that have been passed, try to identify the group number in the corresponding index file

        result = self._generate_index_groups(tmp_dir)
        output = ['echo -e "']
        for param in params.split():
            if param == "or":
                output.append("|")
            elif param == "and":
                output.append("&")
            elif param == "not":
                output.append("!")
            elif param == ";":
                output.append("\n")
            else:
                added_one = False
                for line in result.stdout.split("\n"):
                    parts = line.split()
                    if param in parts and param == parts[1]:
                        idx = parts[0]
                        # print("found index", idx, f"for {param}")
                        added_one = True
                        output.append(idx)
                        break
                if not added_one:
                    output.append(param)
        output.append('\nq"')
        self._logger.log(f"Constructed pipe input {' '.join(output)}", _LE.DEBUG)
        return " ".join(output)

    def _add_index_group(self, tmp_dir, pipe_input):
        ndx_args = [
            "-f",
            os.path.join(tmp_dir, _SGE.STD_STRUCTURE),
            "-o",
            os.path.join(tmp_dir, _SGE.STD_INDEX),
        ]
        self._logger.log(
            f"Added group to index file using command {pipe_input}",
            _LE.DEBUG,
        )
        result = self._backend_executor.execute(
            command=_GE.MAKE_NDX,
            arguments=ndx_args,
            location=tmp_dir,
            check=True,
            pipe_input=self.construct_pipe_arguments(tmp_dir, pipe_input),
        )
        for line in result.stdout.split("\n"):
            self._logger_blank.log(line, _LE.INFO)
        self.get_topol().set_ndx(tmp_dir)



from icolos.utils.enums.step_enums import StepBaseEnum, StepGromacsEnum
from icolos.utils.enums.program_parameters import GromacsEnum
from icolos.core.workflow_steps.gromacs.base import StepGromacsBase
from icolos.utils.execute_external.gromacs import GromacsExecutor
from pydantic import BaseModel
from icolos.core.workflow_steps.step import _LE
import os

_GE = GromacsEnum()
_SGE = StepGromacsEnum()
_SBE = StepBaseEnum


class StepGMXTrjconv(StepGromacsBase, BaseModel):
    """
    Postprocessing step for gromacs trajectories
    Mostly used for removing pbc, fitting trajectory etc.
    """

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=GromacsExecutor)
        self._check_backend_availability()

    def execute(self):

        tmp_dir = self._make_tmpdir()
        topol = self.get_topol()
        topol.write_ndx(tmp_dir)
        for i in range(len(topol.trajectories.keys())):
            topol.write_tpr(tmp_dir, index=i)
            topol.write_trajectory(tmp_dir, index=i)
            fitted_traj = f"traj_{i}.xtc"
            flag_dict = {
                "-s": _SGE.STD_TPR,
                "-f": _SGE.STD_XTC,
                "-o": fitted_traj,
                "-n": _SGE.STD_INDEX,
            }
            arguments = self._parse_arguments(flag_dict=flag_dict)
            result = self._backend_executor.execute(
                command=_GE.TRJCONV,
                arguments=arguments,
                location=tmp_dir,
                check=True,
                pipe_input=self.construct_pipe_arguments(
                    tmp_dir, self.settings.additional[_SBE.PIPE_INPUT]
                ),
            )
            # ensure the old trajectories are not parsed for writeout
            os.remove(os.path.join(tmp_dir, _SGE.STD_XTC))
            for line in result.stdout.split("\n"):
                self._logger_blank.log(line, _LE.DEBUG)
            topol.set_trajectory(path=tmp_dir, file=fitted_traj, index=i)
            topol.set_structure(path=tmp_dir, index=i)
        self._parse_output(tmp_dir)
        self._remove_temporary(tmp_dir)



import os
import json
from typing import Tuple

from icolos.utils.enums.general_utils_enums import JSONSchemasEnum
from icolos.utils.general.files_paths import attach_root_path

_JSE = JSONSchemasEnum()


def _load_schema(path: str) -> dict:
    try:
        with open(path, "r") as f:
            schema_data = f.read()
        schema = json.loads(schema_data)
    except json.JSONDecodeError:
        print(f"Could not load or parse schema file {path}, skipping.")
        schema = {}
    return schema


def _construct_absolute_path(sub_schema: str) -> str:
    rel_path = "src/icolos/config/schemas/"
    return attach_root_path(os.path.join(rel_path, "".join([sub_schema, ".json"])))


def construct_workflow_schema() -> Tuple[dict, str]:
    path = _construct_absolute_path(_JSE.WORKFLOW_SCHEMA)
    return _load_schema(path), os.path.dirname(path)



import functools
import time
from typing import Any
from pydantic import BaseModel


class RetryResult(BaseModel):
    success: bool
    tries: int
    result: Any = None
    exception: Exception = None

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)


# TODO: do a unit test for this
def retry(n_tries, retry_wait_seconds, allowed_exceptions=()):
    if n_tries < 1:
        n_tries = 1

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> RetryResult:
            exc = None
            for idx in range(n_tries):
                try:
                    result = func(*args, **kwargs)
                    return RetryResult(
                        success=True, tries=idx + 1, result=result, exception=None
                    )
                except allowed_exceptions as e:
                    exc = e
                time.sleep(retry_wait_seconds)
            return RetryResult(success=False, tries=n_tries, result=None, exception=exc)

        return wrapper

    return decorator



from copy import deepcopy
from typing import List, Tuple
from icolos.core.workflow_steps.step import StepBase
from pydantic import BaseModel
import tempfile
from icolos.utils.enums.step_enums import StepCressetEnum
from icolos.utils.execute_external.cresset_executor import CressetExecutor
from icolos.utils.general.files_paths import gen_tmp_file
from icolos.core.workflow_steps.step import _LE
import os
from icolos.utils.general.parallelization import Parallelizer, SubtaskContainer


_SCE = StepCressetEnum()


class StepCressetEC(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=CressetExecutor)
        self._check_backend_availability()

    def _prepare_tmp_input(self, batch: List) -> Tuple[List, List]:
        conformers = []
        tmp_dirs = []
        protein = self.data.generic.get_argument_by_extension(
            "pdb", rtn_file_object=True
        )
        for sublist in batch:
            for task in sublist:
                conformer = task.data
                conformers.append(conformer)

                # generate the tmpdir
                tmp_dir = tempfile.mkdtemp()
                tmp_dirs.append(tmp_dir)
                _, path_input_sdf = gen_tmp_file(
                    prefix="tmp_", suffix=".sdf", dir=tmp_dir
                )
                conformer.write(path=path_input_sdf)

                # write the protein to that tmpdir
                protein.write(path=os.path.join(tmp_dir, "protein.pdb"), join=False)

        return conformers, tmp_dirs

    def _execute_cresset_ec_parallel(self):
        parallelizer = Parallelizer(func=self._run_conformer)
        n = 1

        while self._subtask_container.done() is False:

            next_batch = self._get_sublists(
                get_first_n_lists=self._get_number_cores()
            )  # return n lists of length max_sublist_length
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            conformers, tmp_dirs = self._prepare_tmp_input(next_batch)
            self._logger.log(
                f"Executing Cresset EC for batch {n} containing {len(conformers)} conformers",
                _LE.DEBUG,
            )

            parallelizer.execute_parallel(tmp_dir=tmp_dirs, conformer=conformers)

            results = self._parse_results(tmp_dirs, conformers)

            for sublist, result in zip(next_batch, results):
                # TODO: this only works if max length sublist == 1, fine for now as that is all turbomole can handle
                for task in sublist:
                    if result == _SCE.SUCCESS:
                        task.set_status_success()
                    else:
                        task.set_status_failed()
            self._remove_temporary(tmp_dirs)
            n += 1

    def _parse_results(self, tmp_dirs: List, conformers: List):
        # walk over the directory structure, parse the output file, identify the conformer, attach a tag to the mol object
        # TODO: No idea what the output looks like for this, write the parser!!
        pass

    def execute(self):
        # unroll all conformers
        all_conformers = []
        for compound in self.get_compounds():
            for enum in compound.get_enumerations():
                if self._input_object_empty(enum):
                    continue
                else:
                    for conformer in enum.get_conformers():
                        conf = deepcopy(conformer)
                        all_conformers.append(conf)

        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_conformers)
        self._execute_cresset_ec_parallel()

    def _run_conformer(self):
        # run a single conformer through Flare's EC
        self._backend_executor.execute()

    # execution is
    # module load Flare && pyflare electrostaticcomplementarity.py -p protein.pdb ligands.sdf



import shutil
from typing import List
from icolos.core.containers.compound import Compound
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from pydantic import BaseModel
import os
from icolos.utils.enums.step_enums import StepGromacsEnum
from icolos.utils.execute_external.gromacs import GromacsExecutor
from icolos.utils.execute_external.pmx import PMXExecutor
from icolos.utils.enums.program_parameters import (
    GromacsEnum,
    PMXEnum,
    PMXAtomMappingEnum,
    StepPMXEnum,
)
from icolos.utils.general.parallelization import SubtaskContainer

_PE = PMXEnum()
_GE = GromacsEnum()
_SGE = StepGromacsEnum()
_PSE = StepPMXEnum()


class StepPMXabfe(StepPMXBase, BaseModel):
    """Setup files for an ABFE calculation."""

    _gromacs_executor: GromacsExecutor = GromacsExecutor()

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(PMXExecutor)
        self._check_backend_availability()
        self._gromacs_executor = GromacsExecutor(prefix_execution=_SGE.GROMACS_LOAD)

    def execute(self):
        """
        This step manages the setup of a pmx ABFE run for a set of compounds and a protein target.
        Expects:
            - docked compounds provided as an sdf file
            - protein apo structure
            - directory containing mdp files
        """

        assert self.work_dir is not None and os.path.isdir(self.work_dir)
        replicas = (
            self.settings.additional["replicas"]
            if "replicas" in self.settings.additional.keys()
            else 3
        )
        # miror the dir structure for the input files used by the rbfe workflow
        os.makedirs(os.path.join(self.work_dir, "input"), exist_ok=True)
        for folder in ["ligands", "mdp", "protein"]:
            os.makedirs(os.path.join(self.work_dir, "input", folder), exist_ok=True)

        mdp_dir = self.data.generic.get_argument_by_extension(
            ext="mdp", rtn_file_object=True
        )
        # write mdp files to the input dir
        mdp_dir.write(os.path.join(self.work_dir, "input/mdp"))

        # create directory structure
        for comp in self.get_compounds():
            ident = comp.get_index_string()
            os.makedirs(os.path.join(self.work_dir, ident), exist_ok=True)

        # load in the provided apo structure
        protein = self.data.generic.get_argument_by_extension(
            "pdb", rtn_file_object=True
        )
        protein.write(
            os.path.join(self.work_dir, "input/protein/protein.pdb"), join=False
        )
        # parametrise protein
        self._parametrise_protein(
            protein="protein.pdb", path="input/protein", output="protein.gro"
        )
        # self._clean_protein()

        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(self.get_compounds())
        self._execute_pmx_step_parallel(
            run_func=self._parametrise_nodes,
            step_id="parametrize ligands",
            result_checker=self._check_params,
        )

        # now
        self._subtask_container.load_data(self.get_compounds())
        self._execute_pmx_step_parallel(
            run_func=self._setup_abfe,
            step_id="pxm abfe",
            result_checker=self._find_nan_vals,
        )

        # now we make the rest of the dir structure
        for comp in self.get_compounds():
            ident = comp.get_index_string()
            shutil.copyfile(
                os.path.join(self.work_dir, "input/protein/posre.itp"),
                os.path.join(self.work_dir, ident, "complex/posre.itp"),
            )
            for wp in self.therm_cycle_branches:
                wppath = os.path.join(self.work_dir, ident, wp)
                # copy the posre.itp file from input/protein to each

                # stateA/stateB - coupled + decoupled states
                for state in self.states:
                    statepath = os.path.join(wppath, state)
                    os.makedirs(statepath, exist_ok=True)

                    # run1/run2/run3
                    for r in range(1, replicas + 1):
                        runpath = os.path.join(statepath, f"run{r}")
                        os.makedirs(runpath, exist_ok=True)

                        # em/eq_posre/eq/transitions
                        # TODO: this differs from the equil used for RBFE - can we get away without the extra equilibration?
                        for sim in self.settings.additional[_PSE.SIM_TYPES]:
                            simpath = os.path.join(runpath, sim)
                            os.makedirs(simpath, exist_ok=True)

    def _setup_abfe(self, jobs):
        """
        Executes pmx abfe, moves the resulting built files to the right dir
        """
        if isinstance(jobs, list):
            comp = jobs[0]

        args = {
            "-pt": os.path.join(self.work_dir, "input/protein/topol.top"),
            "-lt": os.path.join(
                self.work_dir,
                "input/ligands",
                comp.get_index_string(),
                "MOL.acpype/MOL_GMX.itp",
            ),
            "-pc": os.path.join(self.work_dir, "input/protein/protein.gro"),
            "-lc": os.path.join(
                self.work_dir,
                "input/ligands",
                comp.get_index_string(),
                "MOL.acpype/MOL_GMX.gro",
            ),
            "--build": "",
        }
        self._backend_executor.execute(
            command=_PE.ABFE,
            arguments=self.get_arguments(args),
            location=os.path.join(self.work_dir, comp.get_index_string()),
            check=True,
        )
        # note that this is stochastic, and sometimes it generates bad restraaints/nan values
        # we will simply resubmit n times

    def _find_nan_vals(self, next_batch: List[str]) -> List[List[bool]]:
        """
        Looks throuh the dirs specified in jobs, reads restraints.info
        """
        # sublist length set to 1 for this step
        batch_results = []
        for subtask in next_batch:
            subtask_results = []
            for comp in subtask:
                with open(
                    os.path.join(
                        self.work_dir, comp.get_index_string(), "restraints.info"
                    ),
                    "r",
                ) as f:
                    lines = f.readlines()

                subtask_results.append(any(["nan" in l for l in lines]))
            batch_results.append(subtask_results)
        return batch_results

    def _check_params(self, batch: List[List[Compound]]) -> List[List[bool]]:
        """
        check ligand parameters have been generated properly
        """
        output_files = ["MOL.itp", "MOL.mol2"]
        results = []
        for subjob in batch:
            subjob_results = []
            for job in subjob:
                subjob_results.append(
                    all(
                        [
                            os.path.isfile(
                                os.path.join(self.work_dir, job.get_index_string(), f)
                            )
                            for f in output_files
                        ]
                    )
                )
            results.append(subjob_results)
        return results



from icolos.core.workflow_steps.pmx.atomMapping import StepPMXatomMapping
from icolos.core.workflow_steps.pmx.doublebox import StepPMXdoublebox
from icolos.core.workflow_steps.pmx.genlib import StepPMXgenlib
from icolos.core.workflow_steps.pmx.gentop import StepPMXgentop
from icolos.core.workflow_steps.pmx.ligandHybrid import StepPMXligandHybrid
from icolos.core.workflow_steps.pmx.mutate import StepPMXmutate
from icolos.core.workflow_steps.pmx.abfe import StepPMXabfe
from icolos.core.workflow_steps.pmx.box_water_ions import StepPMXBoxWaterIons
from icolos.core.workflow_steps.pmx.prepare_simulations import StepPMXPrepareSimulations
from icolos.core.workflow_steps.pmx.prepare_transitions import StepPMXPrepareTransitions
from icolos.core.workflow_steps.pmx.run_analysis import StepPMXRunAnalysis
from icolos.core.workflow_steps.pmx.setup_workpath import StepPMXSetup
from icolos.core.workflow_steps.pmx.run_simulations import StepPMXRunSimulations
from icolos.core.workflow_steps.pmx.assemble_systems import StepPMXAssembleSystems



import numpy as np
import pandas as pd

from pydantic import BaseModel
from rdkit.Chem import AllChem
from typing import List

from icolos.core.containers.compound import Conformer

from icolos.core.workflow_steps.step import StepBase
from icolos.utils.enums.step_enums import StepRMSFilterEnum

_SRF = StepRMSFilterEnum()


class StepCalculationBase(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

    def _get_rms_method(self):
        # there are two modes for the execution: "best" is better, but sometimes has performance issues
        # for larger molecules
        if self.settings.additional[_SRF.METHOD] == _SRF.METHOD_ALIGNMOL:
            return AllChem.AlignMol
        elif self.settings.additional[_SRF.METHOD] == _SRF.METHOD_BEST:
            return AllChem.GetBestRMS
        else:
            raise ValueError(
                f"RMS mode {self.settings.arguments.parameters[_SRF.METHOD]} not supported (either {_SRF.METHOD_ALIGNMOL} or {_SRF.METHOD_BEST})."
            )

    @staticmethod
    def _get_property_values(conformers: List[Conformer], prop: str) -> List[float]:
        return [float(conf.get_molecule().GetProp(prop)) for conf in conformers]

    @staticmethod
    def _calculate_rms_matrix(
        conformers: List[Conformer], rms_method, decimals=3
    ) -> pd.DataFrame:
        n_conf = len(conformers)
        df_rms = pd.DataFrame(np.nan, index=range(n_conf), columns=range(n_conf))
        np.fill_diagonal(df_rms.values, 0)

        for i in range(n_conf - 1):
            for j in range(i + 1, n_conf):
                df_rms.iloc[i, j] = df_rms.iloc[j, i] = np.round(
                    rms_method(
                        conformers[i].get_molecule(), conformers[j].get_molecule()
                    ),
                    decimals=decimals,
                )
        return df_rms



import os
import tempfile
from typing import Tuple, List
from copy import deepcopy

from pydantic import BaseModel

from icolos.utils.execute_external.turbomole import TurbomoleExecutor

from icolos.core.containers.compound import Conformer, Enumeration

from icolos.utils.enums.program_parameters import TurbomoleEnum
from icolos.utils.enums.program_parameters import CosmoOutputEnum
from icolos.utils.enums.compound_enums import ConformerContainerEnum
from icolos.core.workflow_steps.calculation.base import StepCalculationBase
from icolos.core.workflow_steps.step import _LE
from icolos.loggers.logger_utils import log_multiline_string
from icolos.utils.general.files_paths import attach_root_path

_EE = TurbomoleEnum()
_CTE = ConformerContainerEnum()
_COE = CosmoOutputEnum()


class StepCosmo(StepCalculationBase, BaseModel):
    """Step that executes Cosmo.

    Note, that the execution (especially in conjunction with a preceding turbomole step) is relatively complex.
    (1) Take the coord file from the additional data attached to the conformers,
    (2) run Cosmo,
    (3) extract the final XYZ snapshot with x2t,
    (4) translate it to a SDF file with obabel and
    (5) combined the new coordinates with the tags."""

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=TurbomoleExecutor)
        self._check_backend_availability()

        # initialize the executor and test availability
        # as they are linked, use a "TurbomoleExecutor" here
        self._initialize_backend(executor=TurbomoleExecutor)
        self._check_backend_availability()

    def _prepare_tmp_input_directories(
        self, enumeration: Enumeration
    ) -> Tuple[List[str], List[str], List[str], List[str]]:
        tmp_dirs = []
        paths_input_cosmofile = []
        paths_config_cosmotherm = []
        paths_output_cosmotherm = []
        for conformer in enumeration:
            # 1) generate all temporary paths
            tmp_dir = tempfile.mkdtemp()
            path_input_cosmofile = os.path.join(tmp_dir, _EE.TM_OUTPUT_COSMOFILE)
            path_config_cosmofile = os.path.join(tmp_dir, _EE.CT_COSMOTHERM_CONFIG_FILE)
            path_output_cosmotherm = os.path.join(
                tmp_dir, _EE.CT_COSMOTHERM_OUTPUT_FILE
            )

            # 2) write-out the COSMO file
            #    Note, that the generation of the COSMO files is part of the Turbomole execution. The reason is, that
            #    the generation is complicated and uses a lot of input form the TM step, thus "cosmoprep" is
            #    executed there.
            if _CTE.EXTRA_DATA_COSMOFILE not in conformer.get_extra_data().keys():
                self._logger.log(
                    f"In order to write out COSMO files, the content needs to be annotated as extra data in the conformers. Have you executed Turbomole before?",
                    _LE.ERROR,
                )
                raise ValueError("Could not find COSMO data to write out - abort.")
            with open(path_input_cosmofile, "w") as f:
                f.writelines(conformer.get_extra_data()[_CTE.EXTRA_DATA_COSMOFILE])

            # 3) add paths
            tmp_dirs.append(tmp_dir)
            paths_input_cosmofile.append(path_input_cosmofile)
            paths_config_cosmotherm.append(path_config_cosmofile)
            paths_output_cosmotherm.append(path_output_cosmotherm)

        return (
            tmp_dirs,
            paths_input_cosmofile,
            paths_config_cosmotherm,
            paths_output_cosmotherm,
        )

    def _execute_run(self, config_path: str):
        result = self._backend_executor.execute(
            command=_EE.CT_COSMOTHERM, arguments=[config_path], check=True
        )
        if _EE.CT_COSMOTHERM_FAIL_STRING in result.stderr:
            self._logger.log(
                f"Execution of {_EE.CT_COSMOTHERM} failed. Error message:", _LE.ERROR
            )
            log_multiline_string(
                logger=self._logger_blank,
                level=_LE.ERROR,
                multi_line_string=result.stdout,
            )

    def _write_config_file(self, config_path: str):
        # by default use the internal configuration, but if one has been specified, use this one
        # note, that the default name of the COSMO file is "mol.cosmo", so this should be used in any config file
        if _EE.CT_CONFIG not in self.settings.arguments.parameters.keys():
            with open(attach_root_path(_EE.CT_CONFIG_DEFAULTPATH), "r") as f:
                config = f.readlines()
            self._logger.log(
                f"Loaded {_EE.CT_COSMOTHERM} configuration from default file {_EE.CT_CONFIG_DEFAULTPATH}.",
                _LE.DEBUG,
            )
        else:
            config = self.settings.arguments.parameters[_EE.CT_CONFIG]
        with open(config_path, "w") as f:
            f.writelines([line.rstrip("\n") + "\n" for line in config])

    def _get_line_by_pattern(self, lines: List[str], pattern: str) -> str:
        for line in lines:
            if pattern in line:
                return line

    def _get_values_from_line(self, line: str) -> List[str]:
        try:
            value_part = line.split(":")[1]
            return value_part.split()
        except Exception:
            return []

    def _annotate_from_output_block(
        self, conformer: Conformer, block: List[str], annotation: dict
    ):
        for key in annotation.keys():
            # get the line with the values
            line = self._get_line_by_pattern(
                lines=block, pattern=annotation[key][_COE.PATTERN]
            )
            if line is None:
                continue

            # get the values and select the one that is to be added
            try:
                values = self._get_values_from_line(line=line)
                value = values[annotation[key][_COE.ELEMENT]]
            except IndexError:
                continue

            # add it as a tag to the conformer; we can replace part of the tag name with e.g. the solvent
            # names if we need to
            conformer.get_molecule().SetProp(key, value)

    def _get_solvents_from_header(self, header: List[str]):
        line_solvents = self._get_line_by_pattern(
            header, pattern=_COE.SOLVENT_BLOCK_HEADER_COMPOUNDS_PATTERN
        )
        return self._get_values_from_line(line_solvents)

    def _get_current_solvent_from_header(self, header: List[str]):
        line_mol_fraction = self._get_line_by_pattern(
            header, pattern=_COE.SOLVENT_BLOCK_HEADER_MOLFRACTION_PATTERN
        )
        solvent_index = self._get_values_from_line(line_mol_fraction).index(
            _COE.SOLVENT_BLOCK_CURRENT_FRACTION_VALUE
        )
        return self._get_solvents_from_header(header)[solvent_index]

    def _parse_general_block(self, lines: List[str], conformer: Conformer):
        general_block = []
        for index in range(len(lines)):
            if _COE.GENERAL_BLOCK_PATTERN_STRING in lines[index]:
                # skip the first lines after the header
                index += 2
                while not lines[index] == "":
                    general_block.append(lines[index])
                    index += 1
                break
        self._annotate_from_output_block(
            conformer=conformer,
            block=general_block,
            annotation=_COE.GENERAL_BLOCK_ANNOTATIONS,
        )

    def _load_solvent_blocks(self, lines: List[str]) -> List[dict]:
        solvent_blocks = []
        index = 0
        while index < len(lines):
            if _COE.SOLVENT_BLOCK_PATTERN_STRING in lines[index]:
                # we need to extract both the header (which solvent?) and the body (actual values)
                new_block = {"header": [], "body": []}
                # go back to start of block
                while (
                    index >= 0 and _COE.SOLVENT_BLOCK_START_PATTERN not in lines[index]
                ):
                    index -= 1

                # extract the header
                while _COE.SOLVENT_BLOCK_BODY_START_PATTERN not in lines[index]:
                    new_block["header"].append(lines[index])
                    index += 1

                # extract the body
                while index < len(lines) and not (
                    lines[index] == "" and lines[index + 1] == ""
                ):
                    new_block["body"].append(lines[index])
                    index += 1
                solvent_blocks.append(new_block)
            index += 1
        return solvent_blocks

    def _annotate_solvent_blocks(
        self, solvent_blocks: List[dict], conformer: Conformer
    ):
        for block_dict in solvent_blocks:
            # get solvent and translate according to internal solvent abbreviation table
            try:
                current_solvent = self._get_current_solvent_from_header(
                    block_dict["header"]
                )
                if current_solvent in _COE.SOLVENT_TRANSLATE_SOLVENT.keys():
                    current_solvent = _COE.SOLVENT_TRANSLATE_SOLVENT[current_solvent]
            except ValueError:
                continue

            # overwrite the solvent name placeholder in and annotate
            template_annotations = deepcopy(_COE.SOLVENT_BLOCK_BODY_ANNOTATIONS)
            annotations = {}
            for key in template_annotations.keys():
                new_key = key.replace(_COE.SOLVENT_REPLACEHOLDER, current_solvent)
                annotations[new_key] = template_annotations[key]

            # annotate
            self._annotate_from_output_block(
                conformer=conformer, block=block_dict["body"], annotation=annotations
            )

    def _parse_output(self, path_output: str, conformer: Conformer):
        # there are two sets of blocks we need to parse: the "general" block, that is always present and, if specified,
        # free energies from solvents ("mixtures")
        # 1) load the file
        with open(path_output, "r") as f:
            lines = f.readlines()
            lines = [line.rstrip("\n") for line in lines]

        # 2) extract the general block: from the match of the pattern line until the second empty line occurs
        #    e.g. "--- Compound 1 (mol) ---\n\nAtomic weights : 111111\n <etc> ...\n\n"
        self._parse_general_block(lines, conformer)

        # 3) extract the solvent blocks (if available)
        #    search for the first occurrence of a Gibb's free energy and expand top until the pattern line is found and
        #    to bottom until more than one empty line is hit; proceed until all blocks are processed
        solvent_blocks = self._load_solvent_blocks(lines)

        if len(solvent_blocks) > 0:
            self._annotate_solvent_blocks(solvent_blocks, conformer)

    def execute(self):
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if len(enumeration.get_conformers()) == 0:
                    continue

                # generate copies of the conformers, as to not accidentally manipulate them
                inp_enum = deepcopy(enumeration)

                # prepare the temporary files and retrieve paths (TM config is charge-state dependent!)
                (
                    tmp_dirs,
                    paths_input_cosmofile,
                    paths_config_cosmotherm,
                    paths_output_cosmotherm,
                ) = self._prepare_tmp_input_directories(enumeration=inp_enum)

                # execute individual conformers
                for (
                    tmp_dir,
                    path_config_cosmotherm,
                    conformer,
                    path_output_cosmotherm,
                ) in zip(
                    tmp_dirs,
                    paths_config_cosmotherm,
                    enumeration.get_conformers(),
                    paths_output_cosmotherm,
                ):
                    self._move_to_dir(tmp_dir)

                    # set a necessary environment variable to avoid clashes
                    os.environ[_EE.TM_TURBOTMPDIR] = tmp_dir

                    # write configuration file
                    self._write_config_file(config_path=path_config_cosmotherm)

                    # all ready; start the execution
                    self._execute_run(config_path=path_config_cosmotherm)

                    # parse the results
                    self._parse_output(
                        path_output=path_output_cosmotherm, conformer=conformer
                    )

                # restore working directory and remove temporary files
                self._restore_working_dir()
                for tmp_dir in tmp_dirs:
                    if os.path.isdir(tmp_dir):
                        self._remove_temporary(tmp_dir)

                self._logger.log(
                    f"Executed COSMO for {len(enumeration.get_conformers())} conformers for enumeration {enumeration.get_index_string()}.",
                    _LE.INFO,
                )



import os
from icolos.core.step_utils.structconvert import StructConvert
from pydantic import BaseModel
from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor
from icolos.utils.enums.step_enums import StepDesmondEnum
from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum

_SDE = StepDesmondEnum()
_SEE = SchrodingerExecutablesEnum()


class StepDesmondSetup(StepSchrodingerBase, BaseModel):
    """
    Run preprocessing step to generate system for Desmond simulation
    """

    _struct_converter: StructConvert = None

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(SchrodingerExecutor)
        self._check_backend_availability()
        self._struct_converter = StructConvert(
            binary_location=self.execution.binary_location,
            prefix_execution=self.execution.prefix_execution,
        )

    def execute(self):
        tmp_dir = self._make_tmpdir()

        # need to take a structure file, possibly preprocess if pdb
        # get the structure file and extract the file name
        structure = self.data.generic.get_argument_by_extension(
            "pdb", rtn_file_object=True
        )
        pdb_id = structure.get_file_name()
        structure.write(tmp_dir)
        # convert the pdb file to mae

        self._struct_converter.pdb2mae(
            os.path.join(tmp_dir, pdb_id),
            os.path.join(tmp_dir, "desmond_md_job_1.mae"),
        )

        defaults = {
            "-HOST": "localhost",
            "-JOBNAME": "desmond_md_job_1",
            "-m": "config.msj desmond_md_job_1.mae",
            "-o": "setup.cms",
        }
        arguments = self._parse_arguments(defaults)
        # compile and write the msj to the tmpdir
        config_dict = (
            self.settings.additional[_SDE.MSJ_FIELDS]
            if _SDE.MSJ_FIELDS in self.settings.additional.keys()
            else {}
        )

        self._write_config(tmp_dir, dict_=config_dict, file_name=_SDE.PREPROCESS_MSJ)

        # execute
        self._backend_executor.execute(
            command=_SEE.MULTISIM_EXEC,
            arguments=arguments,
            check=True,
            location=tmp_dir,
        )
        self._parse_output(tmp_dir)

        self._remove_temporary(tmp_dir)



from icolos.core.workflow_steps.gromacs.cluster import StepGMXCluster
from icolos.core.workflow_steps.gromacs.editconf import StepGMXEditConf
from icolos.core.workflow_steps.gromacs.genion import StepGMXGenion
from icolos.core.workflow_steps.gromacs.grompp import StepGMXGrompp
from icolos.core.workflow_steps.gromacs.mdrun import StepGMXMDrun
from icolos.core.workflow_steps.gromacs.pdb2gmx import StepGMXPdb2gmx
from icolos.core.workflow_steps.gromacs.solvate import StepGMXSolvate
from icolos.core.workflow_steps.gromacs.trjconv import StepGMXTrjconv
from icolos.core.workflow_steps.gromacs.clusters_ts import StepClusterTS
from icolos.core.workflow_steps.gromacs.rsmd import StepGMXrmsd



from typing import List
from icolos.core.containers.compound import Conformer
from icolos.core.workflow_steps.step import StepBase
from pydantic import BaseModel

from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor
from icolos.utils.general.parallelization import Parallelizer
import pandas as pd
import os
from icolos.core.workflow_steps.step import _LE


_SEE = SchrodingerExecutablesEnum




class StepProteinInteraction(StepBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        self._backend_executor = SchrodingerExecutor(
            prefix_execution=self.execution.prefix_execution,
            binary_location=self.execution.binary_location,
        )
        
    def _compute_interactions(self, conf: Conformer):
        tmp_dir = self._make_tmpdir()
        print(tmp_dir)
        conf.write(os.path.join(tmp_dir, "structure.sdf"), format_="pdb")
        command = _SEE.PROTEIN_INTERACTION
        args = self.get_arguments(defaults={"-outfile": "output.csv", "-structure": "structure.pdb"})
        
        
        self._backend_executor.execute(command=command, arguments=args, check=True, location=tmp_dir)
        
        self._parse_output(tmp_dir, conf)
        
        
    def _parse_output(self, path: str, conf: Conformer):
        # parse the output file, attach the boolean to the conformer object
        df = pd.read_csv(os.path.join(path, "output.csv"))
        
        # attach full interaction profile
        conf.add_extra_data("interaction_summary", df)
        
    def _penalize_docking_score(self, conf: Conformer, penalty: float):
        # take the docking score, add a penalty
        docking_score = conf.get_molecule().GetProp("docking_score")
        conf.get_molecule().SetProp("docking_score", str(float(docking_score) + penalty))

    def execute(self):
        """Runs schrodinger's protein_interaction_analysis script"""
        # requies structure file + group 1/2 identifications

        # unroll conformers
        all_confs: List[Conformer] = []
        for comp in self.get_compounds():
            for enum in comp.get_enumerations():
                for conf in enum.get_conformers():
                    all_confs.append(conf)
        
        # attach the interaction information to the conformer
        for conf in all_confs:
            self._compute_interactions(conf)
        
        # attach modified docking score if specific interaction is absent 
        penalty = self._get_additional_setting("penalty", default=1.0)
        for conf in all_confs:
            df = conf.get_extra_data()["interaction_summary"]
            # penalize for every interaction that is not met
            base = self._get_additional_setting("base_residue")
            for interaction in self._get_additional_setting("target_residues"):
                interact_summary = df.loc[df["Residue"].str.contains(base)]["Specific Interactions"]
                print(interact_summary)
                try:
                    if not f"hb to {interaction}" in interact_summary.values[0]:
                        self._logger.log(f"Penalizing docking score for conf {conf.get_index_string()}", _LE.DEBUG)
                        print(f"Penalizing docking score for conf {conf.get_index_string()}", _LE.DEBUG)

                        self._penalize_docking_score(conf, penalty)
                except Exception as e:
                    print(e)
                    # either no interaction summary, or something else went wrong
                    continue
        


        
        
        


import os
import glob
import tempfile
import time
from typing import List, Tuple

from pydantic import BaseModel, Field
from rdkit import Chem
from copy import deepcopy

from icolos.utils.enums.step_enums import StepBaseEnum, StepGoldEnum
from icolos.utils.execute_external.gold import GoldExecutor
from icolos.utils.general.files_paths import gen_tmp_file
from icolos.core.containers.compound import Conformer
from icolos.utils.enums.program_parameters import GoldOutputEnum, GoldEnum
from icolos.core.workflow_steps.step import _LE, StepBase
from icolos.utils.general.parallelization import Subtask, SubtaskContainer, Parallelizer
from icolos.core.step_utils.gold_config import (
    ConfigAutomaticSettings,
    ConfigPopulation,
    ConfigGeneticOperators,
    ConfigFloodFill,
    ConfigDataFiles,
    ConfigFlags,
    ConfigProteinData,
    ConfigFitnessFunctionSettings,
    ConfigSaveOptions,
    ConfigCovalentBonding,
    ConfigConstraints,
    ConfigTermination,
)

_SBE = StepBaseEnum
_CGE = GoldEnum()
_SGE = StepGoldEnum()
_GOE = GoldOutputEnum()


class GoldConfiguration(BaseModel):
    automatic_settings: ConfigAutomaticSettings = Field(
        alias="AUTOMATIC SETTINGS", default=ConfigAutomaticSettings()
    )
    population: ConfigPopulation = Field(alias="POPULATION", default=ConfigPopulation())
    genetic_operators: ConfigGeneticOperators = Field(
        alias="GENETIC OPERATORS", default=ConfigGeneticOperators()
    )
    flood_fill: ConfigFloodFill = Field(alias="FLOOD FILL", default=None)
    data_files: ConfigDataFiles = Field(alias="DATA FILES", default=ConfigDataFiles())
    flags: ConfigFlags = Field(alias="FLAGS", default=ConfigFlags())
    termination: ConfigTermination = Field(
        alias="TERMINATION", default=ConfigTermination()
    )
    constraints: ConfigConstraints = Field(
        alias="CONSTRAINTS", default=ConfigConstraints()
    )
    covalent_bonding: ConfigCovalentBonding = Field(
        alias="COVALENT BONDING", default=ConfigCovalentBonding()
    )
    save_options: ConfigSaveOptions = Field(
        alias="SAVE OPTIONS", default=ConfigSaveOptions()
    )
    fitness_function_settings: ConfigFitnessFunctionSettings = Field(
        alias="FITNESS FUNCTION SETTINGS", default=ConfigFitnessFunctionSettings()
    )
    protein_data: ConfigProteinData = Field(alias="PROTEIN DATA", default=None)


class GoldAdditional(BaseModel):
    configuration: GoldConfiguration = GoldConfiguration()
    gold_config_file: str = None
    docking_score_tag: str = "Gold.Goldscore.Fitness"
    grid_ids: List[str] = ["grid0"]


class StepGold(StepBase, BaseModel):

    gold_additional: GoldAdditional = None
    _gold_executor: GoldExecutor = None

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=GoldExecutor)
        self._check_backend_availability()

        # set Gold specific settings
        self.gold_additional = GoldAdditional(**self.settings.additional)

    def _set_docking_score(self, conformer: Chem.Mol) -> bool:
        tag = self.gold_additional.docking_score_tag
        try:
            docking_score = conformer.GetProp(tag)
        except KeyError:
            self._logger.log(
                f'Could not find tag "{tag}" in conformer, use "docking_score_tag" to adjust.',
                _LE.DEBUG,
            )
            return False
        conformer.SetProp(_SBE.ANNOTATION_TAG_DOCKING_SCORE, str(docking_score))
        return True

    def _generate_temporary_input_output_files(
        self, batch: List[List[Subtask]]
    ) -> Tuple[List[str], List[List[str]], List[List[str]]]:
        tmp_output_dirs = []
        tmp_input_sdf_paths = []
        tmp_output_sdf_paths = []

        for next_subtask_list in batch:
            # generate temporary input files and output directory
            cur_tmp_output_dir = tempfile.mkdtemp()

            # write-out the temporary input files: one molecule per file each
            one_written = False
            list_input_files = []
            list_output_files = []
            for subtask in next_subtask_list:
                enumeration = subtask.data
                mol = deepcopy(enumeration.get_molecule())
                if mol is not None:
                    _, cur_tmp_sdf = gen_tmp_file(suffix=".sdf", dir=cur_tmp_output_dir)
                    writer = Chem.SDWriter(cur_tmp_sdf)
                    mol.SetProp("_Name", enumeration.get_index_string())
                    writer.write(mol)
                    writer.close()
                    list_input_files.append(cur_tmp_sdf)

                    # output files for gold look something like: ranked_tmp1cwtavoh_m1_2.sdf
                    # where: (1) "ranked_" is a constant, (2) followed by the input file name,
                    #        (3) followed by the molecule # (here always "m1" as only 1 compound/file at the moment),
                    #        (4) followed by the pose # and finally (5) the ".sdf" ending
                    # use '*' for getting all files
                    basename_filepart = os.path.basename(cur_tmp_sdf).rstrip(".sdf")
                    list_output_files.append(
                        os.path.join(
                            cur_tmp_output_dir,
                            "".join(["ranked_", basename_filepart, "_m1_*.sdf"]),
                        )
                    )
                    one_written = True
            if one_written is False:
                # no compound left from this batch: remove the temporary folder and skip this one
                self._remove_temporary(cur_tmp_output_dir)
                continue

            # since all went well, add input and output paths here
            tmp_input_sdf_paths.append(list_input_files)
            tmp_output_sdf_paths.append(list_output_files)
            tmp_output_dirs.append(cur_tmp_output_dir)
        return (tmp_output_dirs, tmp_input_sdf_paths, tmp_output_sdf_paths)

    def _execute_gold(self):
        # get number of sublists in batch and initialize Parallelizer
        gold_parallelizer = Parallelizer(func=self._run_subjob)

        # continue until everything is successfully done or number of retries have been exceeded
        while self._subtask_container.done() is False:
            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())

            # generate paths and initialize molecules (so that if they fail, this can be covered)
            (
                tmp_output_dirs,
                tmp_input_sdf_paths,
                tmp_output_sdf_paths,
            ) = self._generate_temporary_input_output_files(next_batch)

            # execute the current batch in parallel; hand over lists of parameters (will be handled by Parallelizer)
            # also increment the tries and set the status to "failed" (don't do that inside subprocess, as data is
            # copied, not shared!)
            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]
            gold_parallelizer.execute_parallel(
                output_dir=tmp_output_dirs, input_paths=tmp_input_sdf_paths
            )

            # parse the output of that particular batch and remove temporary files
            self._parse_gold_output(
                tmp_output_paths=tmp_output_sdf_paths, batch=next_batch
            )

            # clean-up
            self._remove_temporary(tmp_output_dirs)

            # print the progress for this execution
            self._log_execution_progress()

    def _parse_gold_output(
        self, tmp_output_paths: List[List[str]], batch: List[List[Subtask]]
    ):
        def _update_subtask(sublist: List[Subtask], enum_identifier: str):
            for task in sublist:
                if task.data.get_index_string() == enum_identifier:
                    task.set_status_success()

        grid_id = self.gold_additional.grid_ids[0]
        grid_path = self.gold_additional.configuration.protein_data.protein_datafile

        for batch_id in range(len(batch)):
            cur_sublist = batch[batch_id]
            list_comp_output = tmp_output_paths[batch_id]

            for compound_number in range(len(list_comp_output)):

                # expand the output paths to cover all pose files for this compound
                # one input sdf matches to x output sdfs (for x poses)
                comp_sdf_output_paths = glob.glob(list_comp_output[compound_number])

                # loop over all output files
                for comp_output in comp_sdf_output_paths:
                    # this is a protection against the case where empty (file size == 0 bytes) files
                    # are generated due to a failure during docking
                    if (
                        not os.path.isfile(comp_output)
                        or os.path.getsize(comp_output) == 0
                    ):
                        continue

                    mol_supplier = Chem.SDMolSupplier(comp_output, removeHs=False)
                    for mol in mol_supplier:
                        if mol is None:
                            continue

                        # The name is something like: 10:0|tmp8x3rtg4d|sdf|1|dock7
                        cur_enumeration_name = str(mol.GetProp("_Name")).split("|")[0]

                        # add the information on the actual grid used
                        mol.SetProp(_SBE.ANNOTATION_GRID_ID, str(grid_id))
                        mol.SetProp(_SBE.ANNOTATION_GRID_PATH, str(grid_path))
                        mol.SetProp(
                            _SBE.ANNOTATION_GRID_FILENAME, os.path.basename(grid_path)
                        )

                        # if no docking score is attached (i.e. the molecule is a receptor or so, skip it)
                        if self._set_docking_score(mol) is not True:
                            continue

                        # add molecule to the appropriate ligand
                        for compound in self.get_compounds():
                            for enumeration in compound:
                                if (
                                    enumeration.get_index_string()
                                    == cur_enumeration_name
                                ):
                                    new_conformer = Conformer(
                                        conformer=mol,
                                        conformer_id=None,
                                        enumeration_object=enumeration,
                                    )
                                    enumeration.add_conformer(
                                        new_conformer, auto_update=True
                                    )
                                    _update_subtask(
                                        cur_sublist,
                                        enum_identifier=cur_enumeration_name,
                                    )
                                    break

    def _run_subjob(self, output_dir: str, input_paths: List[str]):
        # generate the appropriate config; note that input and output paths are referring to one compound each
        config_path = os.path.join(output_dir, "gold.config")
        self.generate_config_file(path=config_path, ligand_files=input_paths)

        # set up arguments list and execute; change path to temporary sub-folder to avoid cluttering with files
        arguments = [config_path]
        old_dir = os.getcwd()
        os.chdir(output_dir)
        self._backend_executor.execute(
            command=_CGE.GOLD_CALL, arguments=arguments, check=True
        )
        os.chdir(old_dir)
        time.sleep(3)

    def _config_from_file(self, ligand_files: List[str]) -> List[str]:
        config = []
        with open(self.gold_additional.gold_config_file, "r") as f:
            buffer = f.readlines()
            idx = 0
            while idx < len(buffer):
                line = buffer[idx].rstrip("\n")
                config.append(line)
                if _SGE.DATA_FILES in line:
                    # skip over all defined ligand files
                    while _SGE.LIGAND_DATA_FILE in buffer[idx + 1]:
                        self._logger.log(
                            'Skipping pre-defined ligand in configuration file (remove lines starting with "ligand_data_file" to suppress this information).',
                            _LE.DEBUG,
                        )
                        idx = idx + 1

                    # add ligand files as required
                    for ligand_file in ligand_files:
                        config.append(
                            " ".join([_SGE.LIGAND_DATA_FILE, ligand_file, "10"])
                        )
                idx = idx + 1
        return config

    def _config_from_default(self, ligand_files: List[str]) -> List[str]:
        def block_indent(block_name: str) -> str:
            return "".join([_SGE.BLOCK_INDENT, block_name])

        def empty_line() -> str:
            return ""

        def add_block(list_lines: List[str], block: dict):
            for key, value in block.items():
                if key == _SGE.LIGAND_DATA_FILE:
                    # this needs to be overwritten by the actual compound collection
                    self._logger.log(
                        f"Do not set ligand data files explicitly, use the compound handover - skipping.",
                        _LE.WARNING,
                    )
                    continue
                list_lines.append(" ".join([key, "=", str(value)]))
            list_lines.append(empty_line())

        config = [block_indent(_SGE.CONFIGURATION_START), empty_line()]

        # automatic settings
        config.append(block_indent(_SGE.AUTOMATIC_SETTINGS))
        add_block(config, self.gold_additional.configuration.automatic_settings.dict())

        # population
        config.append(block_indent(_SGE.POPULATION))
        add_block(config, self.gold_additional.configuration.population.dict())

        # genetic operators
        config.append(block_indent(_SGE.GENETIC_OPERATORS))
        add_block(config, self.gold_additional.configuration.genetic_operators.dict())

        # flood fill
        config.append(block_indent(_SGE.FLOOD_FILL))
        add_block(config, self.gold_additional.configuration.flood_fill.dict())

        # data files
        config.append(block_indent(_SGE.DATA_FILES))
        for ligand_file in ligand_files:
            config.append(" ".join([_SGE.LIGAND_DATA_FILE, ligand_file, "10"]))
        add_block(config, self.gold_additional.configuration.data_files.dict())

        # flags
        config.append(block_indent(_SGE.FLAGS))
        add_block(config, self.gold_additional.configuration.flags.dict())

        # termination
        config.append(block_indent(_SGE.TERMINATION))
        add_block(config, self.gold_additional.configuration.termination.dict())

        # constraints
        config.append(block_indent(_SGE.CONSTRAINTS))
        add_block(config, self.gold_additional.configuration.constraints.dict())

        # covalent bonding
        config.append(block_indent(_SGE.COVALENT_BONDING))
        add_block(config, self.gold_additional.configuration.covalent_bonding.dict())

        # save options
        config.append(block_indent(_SGE.SAVE_OPTIONS))
        add_block(config, self.gold_additional.configuration.save_options.dict())

        # fitness function settings
        config.append(block_indent(_SGE.FITNESS_FUNCTION_SETTINGS))
        add_block(
            config, self.gold_additional.configuration.fitness_function_settings.dict()
        )

        # protein data
        config.append(block_indent(_SGE.PROTEIN_DATA))
        add_block(config, self.gold_additional.configuration.protein_data.dict())

        return config

    def generate_config_file(self, path: str, ligand_files: List[str]):
        # very complicated custom format for GOLD; see file in "external_documentation/gold.conf" for details

        if self.gold_additional.gold_config_file is None:
            # generate config file step by step from default and overwrite set elements
            # TODO: constraints are not yet supported (and probably a few other fine-grained settings)
            config = self._config_from_default(ligand_files=ligand_files)
        else:
            # load path to config file and replace "ligand_files" as required; as configuration element is ignored in
            # this mode, check that it is not set
            if _SGE.CONFIGURATION in self.gold_additional.dict().keys():
                self._logger.log(
                    'The "configuration" elements are ignored when a config file path is specified.',
                    _LE.WARNING,
                )
            config = self._config_from_file(ligand_files=ligand_files)

        # write out
        with open(path, "w") as f:
            for line in config:
                f.write(line + "\n")

    def execute(self):
        # Note: This step only supports one grid at a time, ensemble docking is taken care of at the workflow level!

        # in order to be able to efficiently execute Gold on the enumeration level, all of them have to be unrolled
        # Note: As they retain their respective Compound object, the attribution later on is simple
        all_enumerations = []
        for compound in self.get_compounds():
            all_enumerations = all_enumerations + compound.get_enumerations()
            for enumeration in compound:
                enumeration.clear_conformers()

        # split into sublists, according to the settings
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_enumerations)

        # execute Gold docking
        self._execute_gold()



from typing import List
import numpy as np
from icolos.loggers.steplogger import StepLogger
from icolos.core.workflow_steps.step import _LE
from scipy.stats import norm
from modAL import ActiveLearner
from skorch.regressor import NeuralNetRegressor
from numpy.linalg import norm
import torch
from torch import nn
from skorch.exceptions import NotInitializedError
from scipy import stats

_logger = StepLogger()


class FeedForwardNet(nn.Module):
    def __init__(self) -> None:
        super(FeedForwardNet, self).__init__()

        self.lin1 = nn.Linear(2048, 256)
        self.relu1 = nn.LeakyReLU()
        self.lin2 = nn.Linear(256, 256)
        self.relu2 = nn.LeakyReLU()
        self.lin3 = nn.Linear(256, 256)
        self.relu3 = nn.LeakyReLU()
        self.dropout = nn.Dropout()
        self.lin4 = nn.Linear(256, 1)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        x = self.relu1(self.lin1(inputs))
        x = self.relu2(self.lin2(x))
        x = self.relu3(self.lin3(x))
        x = self.lin4(x)
        return x

    def get_latent_embedding(self, inputs: torch.Tensor) -> torch.Tensor:
        """Return the latent feature representatino from the last hidden layer

        :param torch.Tensor inputs: input features
        :return _type_: _description_
        """
        with torch.no_grad():
            x = self.relu1(self.lin1(inputs))
            x = self.relu2(self.lin2(x))
            x = self.lin3(x)
        return x


def latent_distance(
    learner: ActiveLearner,
    X: np.ndarray,
    previous_idx: List[int],
    n_instances: int,
    *args,
    **kwargs,
):
    """Selects new points to acquire based on the UQ approach detailed here https://doi.org/10.1039/C9SC02298H

    :param ActiveLearner learner: trained active learning model to query
    :param np.ndarray X: array containing the fingerprints to predict values from
    """
    try:
        predictions = learner.predict(X)

    except NotInitializedError:
        _logger.log("estimator not fitted, selecting random compounds ...", _LE.DEBUG)

        indices = range(0, X.shape[0])
        return np.random.choice(indices, replace=False, size=n_instances)

    # compute pairwise euclidian distances of each predicted point to the known points, return those indices with the highest distance to known
    known_X = X[previous_idx]

    # access the skorch object
    estimator: NeuralNetRegressor = learner.estimator
    known_latent_embedding = estimator.module_.get_latent_embedding(
        inputs=torch.from_numpy(known_X).to(estimator.device)
    )
    # n_samples, emb_dim
    # include the full dataset here to retain indexing, some norms will be 0, but that is fine
    all_embedding = estimator.module_.get_latent_embedding(
        inputs=torch.from_numpy(X).to(estimator.device)
    )
    known_latent_embedding, all_embedding = (
        known_latent_embedding.cpu(),
        all_embedding.cpu(),
    )

    # now compute distance to each known embedding vector for each unknown point
    distances = []
    # print("prev_idx", previous_idx)
    zero_count = 0
    for test_embedding in all_embedding:
        # distances_to_known = []
        # for known_emb in known_latent_embedding:
        #     # min(abs(x), axis=0)
        #     # return the distance to the closest point in the training set
        #     # Note, these distances are uncalibrated
        #     # embeddings of dim = 128
        diff = test_embedding - known_latent_embedding
        distances.append(np.linalg.norm(diff))

        #     distances_to_known.append(distance.euclidean(test_embedding, known_emb))
        # # for now, simply use the minimum distance to a known example as the similarity metric
        # if min(distances_to_known) == 0:
        #     zero_count += 1
        # distances.append(np.mean((distances_to_known)))

    print(f"got {zero_count} zero-distances for {len(previous_idx)} compounds")

    sorted_preds = np.argsort(np.array(distances)).reshape(-1)

    sorted_preds = sorted_preds[zero_count : n_instances + zero_count]

    # distances corresponding to selected indices
    # final_indices = []
    # for index in sorted_preds:
    #     final_indices.append(distances[index])

    return sorted_preds


def greedy_acquisition(
    learner: ActiveLearner,
    X: np.ndarray,
    previous_idx: List[int],
    n_instances: int,
    warmup: bool = False,
    epsilon: float = 0.0,
) -> np.ndarray:
    """Implements greedy acquisition by querying model for top predicted points, note that typically we are dealing with affinity prediction so highest is not best.

    :param  estimator: SKLearn-type estimator to be queried
    :param np.ndarray X: array of fingerprints for each compound to be predicted by the model
    :param List[int] previous_idx: List of the previously queried indices to avoid repitition
    :param int n_instances: batch size to query
    :param bool warmup: Control warmup period in which random samples are generated, defaults to False
    :param float epsilon: enable greedy epsilon acquisition, defaults to 0.0
    :return np.ndarray: array of top predictions from the estimator
    """
    if warmup:
        _logger.log("Warmup epoch, using random sampling...", _LE.DEBUG)
        indices = range(0, X.shape[0])
        # sample indices without replacement
        return np.random.choice(indices, replace=False, size=n_instances)
    else:
        try:
            predictions = learner.predict(X)

        except:
            _logger.log(
                "estimator not fitted, selecting random compounds ...", _LE.DEBUG
            )

            indices = range(0, X.shape[0])
            return np.random.choice(indices, replace=False, size=n_instances)

    # zero those predictions we've seen before
    # TODO: remove this, just predict on the unseen variables
    for idx in previous_idx:
        predictions[idx] = 0
    # smaller before n_instances, largest after, take the most negative
    sorted_preds = np.argpartition(predictions, n_instances, axis=0).reshape(-1)[
        :n_instances
    ]

    if epsilon > 0.0:
        # replace that fraction of the predictions with random samples
        n_replacements = int(n_instances * epsilon)

        # select random indices in sorted preds to select
        indices_to_replace = np.random.choice(
            [i for i in range(len(sorted_preds))], n_replacements, replace=False
        )
        # select random compound incides from the dataset to substitute into the returned predictions
        replacements = [np.random.randint(0, X.shape[0]) for _ in range(n_replacements)]
        _logger.log(
            f"Replacing {len(indices_to_replace)} predictions with random values",
            _LE.DEBUG,
        )
        for idx, val in zip(indices_to_replace, replacements):
            predictions[idx] = val
    return sorted_preds


def expected_improvement(
    learner: ActiveLearner,
    X: np.ndarray,
    previous_idx: List[int],
    n_instances: int,
    warmup: bool = False,
    highest_is_best: bool = False,
    **kwargs,
) -> np.ndarray:
    """Select new points to acquire based on expected improvement

    :param  estimator: SKLearn-type estimator to be queried
    :param np.ndarray X: array of fingerprints for each compound to be predicted by the model
    :param List[int] previous_idx: List of the previously queried indices to avoid repitition
    :param int n_instances: batch size to query
    :param bool warmup: Control warmup period in which random samples are generated, defaults to False
    :param float epsilon: enable greedy epsilon acquisition, defaults to 0.0
    :return np.ndarray: array of top predictions from the estimator
    """

    if warmup:
        _logger.log("Warmup epoch, using random sampling...", _LE.DEBUG)
        indices = range(0, X.shape[0])
        return np.random.choice(indices, replace=False, size=n_instances)

    estimator = learner.estimator
    if not isinstance(estimator, NeuralNetRegressor):
        subestimates = []
        estimators = estimator.estimators_

        for i in range(len(estimators)):
            estimator.estimators_ = estimators[0 : i + 1]
            subestimates.append(estimator.predict(X))
        subestimates = np.array(subestimates)
        stdev = np.std(subestimates, axis=0).reshape((-1))
    # for neural networks, get the calibrated uncertainties from latent distance metrics
    else:
        known_X = X[previous_idx]

        # access the skorch object
        estimator: NeuralNetRegressor = learner.estimator
        known_latent_embedding = estimator.module_.get_latent_embedding(
            inputs=torch.from_numpy(known_X).to(estimator.device)
        )
        # n_samples, emb_dim
        # include the full dataset here to retain indexing
        all_embedding = estimator.module_.get_latent_embedding(
            inputs=torch.from_numpy(X).to(estimator.device)
        )
        known_latent_embedding, all_embedding = (
            known_latent_embedding.cpu(),
            all_embedding.cpu(),
        )

        # now compute distance to each known embedding vector for each unknown point
        stdev = []
        for test_embedding in all_embedding:
            diff = test_embedding - known_latent_embedding
            stdev.append(np.linalg.norm(diff))

    y_hat = np.array(learner.predict(X))
    mu = np.mean(y_hat)
    y_best = np.max(y_hat) if highest_is_best else np.min(y_hat)
    gamma = (mu - y_best) / stdev
    ei = stdev * gamma * stats.norm.cdf(gamma) + stdev * stats.norm.pdf(gamma)

    for idx in previous_idx:
        ei[idx] = 0
    # smaller before n_instances, largest after
    return np.argpartition(ei, n_instances, axis=0).reshape(-1)[:n_instances]



from rdkit.Chem import Mol
from rdkit.Chem.rdMolDescriptors import CalcNumRings, CalcNumAromaticRings
from pydantic import BaseModel

from icolos.utils.enums.program_parameters import FeatureCounterEnum
from icolos.utils.enums.step_enums import StepFeatureCounterEnum
from icolos.core.workflow_steps.step import _LE
from icolos.core.workflow_steps.calculation.base import StepCalculationBase

_FC = FeatureCounterEnum()
_SFC = StepFeatureCounterEnum()


class StepFeatureCounter(StepCalculationBase, BaseModel):
    def __init__(self, **data):
        super().__init__(**data)

        # extend parameters with defaults
        if _SFC.LEVEL not in self.settings.additional.keys():
            self.settings.additional[_SFC.LEVEL] = _SFC.LEVEL_CONFORMER
            self._logger.log(
                f'No operational level for feature counting specified, defaulting to "{_SFC.LEVEL_CONFORMER}".',
                _LE.INFO,
            )

    def _count_rings(self, mol: Mol):
        number_rings = CalcNumRings(mol)
        mol.SetProp(_FC.PROPERTY_NUM_RINGS, str(number_rings))

    def _count_aromatic_rings(self, mol: Mol):
        number_rings = CalcNumAromaticRings(mol)
        mol.SetProp(_FC.PROPERTY_NUM_AROMATIC_RINGS, str(number_rings))

    def _get_feature_method(self, feature: str):
        if feature == _FC.PROPERTY_NUM_RINGS:
            return self._count_rings
        elif feature == _FC.PROPERTY_NUM_AROMATIC_RINGS:
            return self._count_aromatic_rings
        else:
            raise ValueError(f'Feature "{feature}" not yet supported.')

    def execute(self):
        feature = self.settings.additional[_SFC.FEATURE].lower()
        feature_method = self._get_feature_method(feature=feature)
        level = self.settings.additional[_SFC.LEVEL]
        mol_count = 0
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if level == _SFC.LEVEL_ENUMERATION:
                    mol = enumeration.get_molecule()
                    if mol is not None:
                        feature_method(mol)
                        mol_count = mol_count + 1
                elif level == _SFC.LEVEL_CONFORMER:
                    for conformer in enumeration.get_conformers():
                        mol = conformer.get_molecule()
                        if mol is not None:
                            feature_method(mol)
                            mol_count = mol_count + 1
                else:
                    raise ValueError(f'Level "{level}" not supported.')
        self._logger.log(
            f'Counted feature "{feature}" for {mol_count} molecules.', _LE.INFO
        )



import os
import shutil
import time
import json
import tempfile
from typing import Tuple

from icolos.utils.enums.step_enums import StepBaseEnum
from icolos.utils.enums.general_utils_enums import CheckFileGenerationEnum

_SE = StepBaseEnum
_FG = CheckFileGenerationEnum()


def check_file_availability(
    path: str, interval_sec: int = 1, maximum_sec: int = 10
) -> str:
    counter = 0
    while not os.path.exists(path):
        # wait for an interval
        time.sleep(interval_sec)
        counter = counter + 1

        # if there's time left, proceed
        if maximum_sec is not None and (counter * interval_sec) > maximum_sec:
            break
    if os.path.exists(path):
        if os.path.getsize(path) == 0:
            return _FG.GENERATED_EMPTY
        else:
            return _FG.GENERATED_SUCCESS
    else:
        return _FG.NOT_GENERATED


def remove_folder(folder_path: str):
    if os.path.isdir(folder_path):
        shutil.rmtree(folder_path)


def empty_output_dir(path: str):
    for root, subf, files in os.walk(path):
        for file in files:
            os.remove(os.path.join(root, file))


def move_up_directory(path, n=1):
    """Function, to move up 'n' directories for a given "path"."""
    # add +1 to take file into account
    if os.path.isfile(path):
        n += 1
    for _ in range(n):
        path = os.path.dirname(os.path.abspath(path))
    return path


def attach_root_path(path, n=4):
    """Function to attach the root path of the module for a given "path"."""
    ROOT_DIR = move_up_directory(os.path.abspath(__file__), n=n)
    return os.path.join(ROOT_DIR, path)


def lines_in_file(path):
    with open(path) as f:
        for i, l in enumerate(f):
            pass
    return i + 1


def dict_from_json_file(path):
    with open(path, "r") as f:
        return json.load(f)


def any_in_file(path, strings):
    if isinstance(strings, str):
        strings = [strings]
    if os.path.isfile(path):
        with open(path, "r") as f:
            file_raw = f.readlines()
            for string in strings:
                if any(string in line for line in file_raw):
                    return True
            return False
    else:
        return False


def infer_input_type(path: str) -> str:
    basename = os.path.basename(path)
    ending = basename[-3:].upper()
    if ending in [_SE.FORMAT_SDF, _SE.FORMAT_CSV, _SE.FORMAT_SMI]:
        return ending
    else:
        raise ValueError(f"Ending {ending} not supported.")


def gen_tmp_file(
    suffix: str = None, prefix: str = None, dir: str = None, text: bool = True
) -> Tuple[str, str]:
    """Function wraps tempfile.mkstemp(), but closes the connection and returns the file name instead of the handler."""
    # note that in contrast to the underlying "mkstemp" function, "text" is set to True here
    fhandle, path = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dir, text=text)
    os.close(fhandle)
    return os.path.basename(path), path



import os
from copy import deepcopy
from tempfile import mkdtemp
from typing import List, Tuple
from pydantic import BaseModel
from rdkit import Chem

from icolos.core.containers.compound import Conformer
from icolos.core.workflow_steps.calculation.base import StepCalculationBase
from icolos.utils.enums.program_parameters import KallistoEnum, OpenBabelEnum
from icolos.utils.enums.step_enums import StepKallistoEnum
from icolos.core.workflow_steps.step import _LE
from icolos.utils.execute_external.kallisto import KallistoExecutor
from icolos.utils.execute_external.openbabel import OpenBabelExecutor
from icolos.utils.general.files_paths import gen_tmp_file
from icolos.utils.general.icolos_exceptions import StepFailed
from icolos.utils.general.parallelization import SubtaskContainer, Parallelizer

_SKE = StepKallistoEnum()
_KE = KallistoEnum()
_OBE = OpenBabelEnum()

_all_kallisto_commands = [
    _KE.ALP,
    _KE.BONDS,
    _KE.CNS,
    _KE.EEQ,
    _KE.EXS,
    _KE.LIG,
    _KE.PROX,
    _KE.RMS,
    _KE.SORT,
    _KE.STM,
    _KE.VDW,
]


class KallistoAdditional(BaseModel):
    features: List[str] = [_KE.ALP, _KE.BONDS]  # list of features to be obtained


class StepKallisto(StepCalculationBase, BaseModel):

    kallisto_additional: KallistoAdditional = None
    _openbabel_executor: OpenBabelExecutor = None

    def __init__(self, **data):
        super().__init__(**data)

        self._initialize_backend(executor=KallistoExecutor)
        self._check_backend_availability()

        # initialize the executor for all "OpenBabel"
        self._openbabel_executor = OpenBabelExecutor()
        if not self._openbabel_executor.is_available():
            raise StepFailed(
                "Kallisto requires OpenBabel execution, initialization failed."
            )

        # initialize the additional settings
        self.kallisto_additional = KallistoAdditional(**self.settings.additional)

    def _prepare_temp_input(self, tmp_dir: str, molecule: Chem.Mol) -> str:
        tmp_sdf_path = super()._prepare_temp_input(tmp_dir, molecule)

        # Kallisto expects the input either in a mol2 file or an XYZ file
        _, tmp_xyz_path = gen_tmp_file(suffix=".xyz", dir=tmp_dir)

        # translate the
        arguments = [
            tmp_sdf_path,
            _OBE.OBABEL_INPUTFORMAT_SDF,
            _OBE.OBABEL_OUTPUT_FORMAT_XYZ,
            "".join([_OBE.OBABEL_O, tmp_xyz_path]),
        ]
        self._openbabel_executor.execute(
            command=_OBE.OBABEL, arguments=arguments, check=False
        )

        self._logger.log(
            f"Translated input molecule to file {tmp_xyz_path}.", _LE.DEBUG
        )
        return tmp_xyz_path

    def _prepare_batch(self, batch) -> Tuple:
        tmp_dirs = []
        input_files = []
        output_files = []
        conformers = []
        for next_subtask_list in batch:
            tmp_dir = mkdtemp()
            tmp_dirs.append(tmp_dir)
            for subtask in next_subtask_list:
                _, tmp_out_path = gen_tmp_file(suffix=".out", dir=tmp_dir)
                output_files.append(tmp_out_path)
                conformer = subtask.data
                conformers.append(conformer)
                input_xyz_file = self._prepare_temp_input(
                    tmp_dir, conformer.get_molecule()
                )
                input_files.append(input_xyz_file)
        return tmp_dirs, input_files, output_files, conformers

    def _prepare_arguments(self, settings: List) -> List:

        # add flags
        for flag in self.settings.arguments.flags:
            settings.append(flag)

        # add parameters
        parameters = deepcopy(self.settings.arguments.parameters)

        # flatten the dictionary into a list for command-line execution
        for key in parameters.keys():
            if key in _all_kallisto_commands:
                self._logger.log(
                    f"Use the additional block to specify Kallisto commands, parameter {key} ignored.",
                    _LE.WARNING,
                )
                continue
            settings.append(key)
            settings.append(parameters[key])
        return settings

    def _run_subjob(self, tmp_dir: str, input_file: str, output_file: str) -> None:
        work_dir = os.getcwd()
        os.chdir(tmp_dir)

        # construct the specified command-line call
        arguments = []
        for feature in self.kallisto_additional.features:
            if feature not in _all_kallisto_commands:
                self._logger.log(
                    f"Kallisto feature {feature} not supported, will be ignored.",
                    _LE.WARNING,
                )
                continue
            arguments.append(feature)
            arguments.append(input_file)
        arguments = self._prepare_arguments(arguments)

        result = self._backend_executor.execute(
            command=_KE.KALLISTO, arguments=arguments, check=False
        )

        # Kallisto prints the result to stdout -> store it in a temporary file
        with open(output_file, "w") as f:
            f.writelines(result.stdout)

        os.chdir(work_dir)

    def _parse_kallisto_result(
        self, output_files: List[str], conformers: List[Conformer]
    ) -> List:
        def _split_list(inp: List, chunk_size: int) -> List[List[str]]:
            assert len(inp) % chunk_size == 0
            return [
                inp[idx : idx + chunk_size] for idx in range(0, len(inp), chunk_size)
            ]

        results = []
        number_features = len(self.kallisto_additional.features)

        for output_file, conformer in zip(output_files, conformers):
            number_atoms = conformer.get_molecule().GetNumAtoms()

            # load features from output file
            with open(output_file, "r") as f:
                features_lines = f.readlines()
            features_lines = [line.lstrip().rstrip() for line in features_lines]

            # check, that all features could be calculated for all atoms
            expected_lines = number_features * number_atoms
            if expected_lines != len(features_lines):
                self._logger.log(
                    f"Kallisto result for conformer {conformer.get_index_string()} incomplete ({len(features_lines)} lines instead of the expected {expected_lines}), check {output_file} - proceeding.",
                    _LE.WARNING,
                )
                results.append(_SKE.FAILURE)
                continue

            # group the features and add them to the conformers
            sublists = _split_list(features_lines, number_atoms)
            for feature, sublist in zip(self.kallisto_additional.features, sublists):
                conformer.get_molecule().SetProp(feature, "|".join(sublist))
            results.append(_SKE.SUCCESS)
        return results

    def _execute_kallisto(self):
        kallisto_parallelizer = Parallelizer(func=self._run_subjob)
        n = 1
        while self._subtask_container.done() is False:
            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())
            tmp_dirs, input_files, output_files, conformers = self._prepare_batch(
                next_batch
            )

            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            self._logger.log(f"Executing Kallisto for batch {n}.", _LE.DEBUG)

            kallisto_parallelizer.execute_parallel(
                tmp_dir=tmp_dirs, input_file=input_files, output_file=output_files
            )

            results = self._parse_kallisto_result(output_files, conformers)
            for sublist, result in zip(next_batch, results):
                assert len(sublist) == 1
                for task in sublist:
                    if result == _SKE.SUCCESS:
                        task.set_status_success()
                    else:
                        task.set_status_failed()
            n += 1
            self._remove_temporary(tmp_dirs)

    def execute(self):
        assert len(self.kallisto_additional.features) > 0
        all_conformers = []
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if enumeration.get_conformers():
                    for conformer in enumeration.get_conformers():
                        all_conformers.append(conformer)
        self.execution.parallelization.max_length_sublists = 1
        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_conformers)
        self._execute_kallisto()
        self._logger.log(
            f"Completed execution of Kallisto for {len(all_conformers)} conformers, attached the following features: [{', '.join(self.kallisto_additional.features)}].",
            _LE.INFO,
        )



import os

from pydantic import BaseModel, PrivateAttr
from rdkit import Chem
from copy import deepcopy

from icolos.core.workflow_steps.schrodinger.base import StepSchrodingerBase
from icolos.utils.execute_external.prime import PrimeExecutor
from icolos.utils.execute_external.schrodinger import SchrodingerExecutor

from icolos.core.containers.compound import Conformer

from icolos.utils.enums.program_parameters import PrimeEnum, SchrodingerExecutablesEnum
from icolos.utils.enums.step_enums import StepPrimeEnum, StepGlideEnum
from icolos.core.workflow_steps.step import _LE
from icolos.core.step_utils.sdconvert_util import SDConvertUtil
from icolos.core.step_utils.structcat_util import StructcatUtil
from icolos.utils.general.files_paths import gen_tmp_file
from icolos.utils.general.parallelization import SubtaskContainer, Parallelizer
from tempfile import mkdtemp

_SPE = StepPrimeEnum()
_PE = PrimeEnum()
_SEE = SchrodingerExecutablesEnum()
_SGE = StepGlideEnum()


class StepPrime(StepSchrodingerBase, BaseModel):
    """
    Interface to Schrodinger's Prime mmgbsa implementation
    """

    _schrodinger_executor: SchrodingerExecutor = None

    class Config:
        underscore_attrs_are_private = True

    _sdconvert_util = PrivateAttr()
    _structcat_util = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)

        # initialize the executor and test availability
        self._initialize_backend(executor=PrimeExecutor)
        self._check_backend_availability()
        self._schrodinger_executor = SchrodingerExecutor(
            prefix_execution=self.execution.prefix_execution
        )

        # prepare sdconvert utility
        self._sdconvert_util = SDConvertUtil(
            prefix_execution=self.execution.prefix_execution,
            binary_location=self.execution.binary_location,
        )

        # prepare structcat utility
        self._structcat_util = StructcatUtil(
            prefix_execution=self.execution.prefix_execution,
            binary_location=self.execution.binary_location,
        )

    def _execute_prime(self):
        # note, that as the output file name cannot be set (an "-out.maegz" will be attached), this does
        # not need to be heeded here and is encoded in the fixed file name strings

        prime_parallelizer = Parallelizer(func=self._run_subjob)
        n = 1

        while self._subtask_container.done() is False:

            next_batch = self._get_sublists(get_first_n_lists=self._get_number_cores())

            # generate lists for the next batch
            tmp_dirs, complex_paths, output_sdf_paths = self._prepare_batch(next_batch)

            self._apply_token_guard()

            _ = [sub.increment_tries() for element in next_batch for sub in element]
            _ = [sub.set_status_failed() for element in next_batch for sub in element]

            self._logger.log(f"Executing prime for batch {n}", _LE.DEBUG)

            prime_parallelizer.execute_parallel(
                complex_path=complex_paths,
                sdf_output=output_sdf_paths,
                tmp_output_dir=tmp_dirs,
            )

            self._parse_prime_output(
                complex_paths, tmp_dirs, output_sdf_paths, next_batch
            )
            n += 1

    def _parse_prime_output(self, complex_paths, tmp_dirs, output_sdf_paths, batch):
        # go through the batch, get the info from the output file and
        scores = []
        for i in range(len(output_sdf_paths)):
            sdf_path = output_sdf_paths[i]
            curr_enum = None
            curr_conformer = None
            mol_supplier = Chem.SDMolSupplier(sdf_path, removeHs=False)
            for mol in mol_supplier:
                # check whether the name corresponds to an enum or conformer
                identifier = str(mol.GetProp("_Name"))
                is_enum = True if len(identifier.split(":")) == 2 else False
                if (
                    not is_enum
                ):  # if we are dealing with a conformer, drop the conformer index to get the enum id
                    enum_index = ":".join(list(mol.GetProp("_Name").split(":"))[:-1])
                else:
                    enum_index = identifier
                # extract the enumeration object, regardless of whether we're dealing with a conformer or an enumeration
                prime_score = mol.GetProp(_SPE.MMGBSA_SCORE)
                for compound in self.get_compounds():
                    for enumeration in compound.get_enumerations():
                        if enumeration.get_index_string() == enum_index:
                            curr_enum = enumeration
                # if we have a conformer, find the conformer with the right ID and append the score to the existing object
                if not is_enum:
                    assert curr_enum is not None
                    for conformer in curr_enum.get_conformers():
                        if conformer.get_index_string() == identifier:
                            curr_conformer = conformer

                else:
                    # if scoring an enumeration, create and attach a conformer out of it
                    curr_conformer = Conformer(conformer=mol)

                    curr_enum.add_conformer(curr_conformer)

                assert curr_conformer is not None
                # now we have the conformer from the originals, set the prime score
                curr_conformer.get_molecule().SetProp(_SPE.MMGBSA_SCORE, prime_score)
                self._logger.log(
                    f"Calculated dG Bind of {prime_score} for conformer {curr_conformer.get_index_string()}",
                    _LE.INFO,
                )
                scores.append(prime_score)

        # after parsing, remove the directories
        self._remove_temporary(tmp_dirs)

        # set success status
        for sublist in batch:
            for task in sublist:
                task.set_status_success()

    def _prepare_batch(self, batch):
        # generate input files for the batch and return tmpdirs

        tmp_dirs = []
        complex_paths = []
        output_sdf_paths = []
        for next_subtask_list in batch:
            tmp_dir = mkdtemp()
            _, tmp_input_sdf_file = gen_tmp_file(suffix=".sdf", dir=tmp_dir)
            _, tmp_input_mae_file = gen_tmp_file(suffix=".maegz", dir=tmp_dir)
            _, tmp_output_sdf_file = gen_tmp_file(suffix=".sdf", dir=tmp_dir)
            writer = Chem.SDWriter(tmp_input_sdf_file)
            for subtask in next_subtask_list:
                mol = deepcopy(subtask.data.get_molecule())
                conf_id = subtask.data.get_index_string()
                mol.SetProp("_Name", conf_id)
                writer.write(mol)
            writer.close()

            # now we have an sdf file with all the conformers from that batch.  Attach the
            structcat_args = [
                "-imae",
                self.settings.additional[_SPE.RECEPTOR],
                "-isd",
                tmp_input_sdf_file,
                "-omae",
                tmp_input_mae_file,
            ]
            self._schrodinger_executor.execute(
                command=_SEE.STRUCTCAT,
                arguments=structcat_args,
                location=tmp_dir,
                check=True,
            )

            tmp_dirs.append(tmp_dir)
            complex_paths.append(tmp_input_mae_file)
            output_sdf_paths.append(tmp_output_sdf_file)

        return tmp_dirs, complex_paths, output_sdf_paths

    def _run_subjob(self, complex_path, sdf_output, tmp_output_dir):

        work_dir = os.getcwd()
        os.chdir(tmp_output_dir)

        arguments = [complex_path, _PE.PRIME_OUTTYPE, _PE.PRIME_OUTTYPE_LIGAND]
        for key in self.settings.arguments.parameters.keys():
            if key not in [_PE.PRIME_OUTTYPE, _PE.PRIME_NJOBS]:
                arguments.append(key)
                arguments.append(str(self.settings.arguments.parameters[key]))
        for flag in self.settings.arguments.flags:
            arguments.append(str(flag))
        if _PE.PRIME_WAIT not in arguments:
            arguments.append(_PE.PRIME_WAIT)

        result = self._backend_executor.execute(
            command=_PE.PRIME_MMGBSA,
            arguments=arguments,
            check=True,
            location=tmp_output_dir,
        )
        print(result)

        output_file = complex_path.split(".")[0] + "-out.maegz"
        assert os.path.isfile(output_file)
        # Convert the mae ligand output back to sdf
        self._sdconvert_util.mae2sdf(output_file, sdf_output)
        os.chdir(work_dir)
        return result

    def execute(self):
        # need to unwrap the conformers to efficiently run in parallel, create lists of subtasks, each with their own files, tmpdirs, then execute them in parallel
        all_conformers = []
        for compound in self.get_compounds():
            for enumeration in compound.get_enumerations():
                if enumeration.get_conformers():
                    # default running mode is to score incoming conformers without changing their configurations
                    for conformer in enumeration.get_conformers():
                        all_conformers.append(conformer)
                else:
                    all_conformers.append(enumeration)

        self._subtask_container = SubtaskContainer(
            max_tries=self.execution.failure_policy.n_tries
        )
        self._subtask_container.load_data(all_conformers)
        self._execute_prime()
        self._logger.log(
            f"Executed Prime for {len(all_conformers)} confomers", _LE.DEBUG
        )



from icolos.utils.enums.program_parameters import SchrodingerExecutablesEnum
from icolos.utils.execute_external.execute import ExecutorBase

SEE = SchrodingerExecutablesEnum()


class SDConvertExecutor(ExecutorBase):
    """For the execution of the "sdconvert" binary."""

    def __init__(self, prefix_execution=None, binary_location=None):
        super().__init__(
            prefix_execution=prefix_execution, binary_location=binary_location
        )

    def execute(
        self, command: str, arguments: list, check=True, location=None, pipe_input=None
    ):
        # check, whether a proper executable is provided
        if command not in [SEE.SDCONVERT]:
            raise ValueError(
                "Parameter command must be an dictionary of the internal sdconvert executable list."
            )

        # take care of the special path to "sdconvert"
        if command == SEE.SDCONVERT:
            command = SEE.SDCONVERT_CALL

        return super().execute(
            command=command,
            arguments=arguments,
            check=check,
            location=None,
            pipe_input=pipe_input,
        )

    def is_available(self):
        try:
            # raise NotImplementedError
            result = self.execute(
                command=SEE.SDCONVERT, arguments=SEE.SDCONVERT_HELP, check=False
            )

            if SEE.SDCONVERT_HELP_IDENTIFICATION_STRING in result.stdout:
                return True
            return False
        except Exception as e:
            return False



from icolos.core.containers.perturbation_map import PerturbationMap
from icolos.core.workflow_steps.pmx.base import StepPMXBase
from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.program_parameters import PMXEnum
from icolos.utils.execute_external.pmx import PMXExecutor
from pydantic import BaseModel
import os

_PE = PMXEnum()
_LE = LoggingConfigEnum()


class StepPMXgentop(StepPMXBase, BaseModel):
    """Fill hybrid topology with B states."""

    def __init__(self, **data):
        super().__init__(**data)
        self._initialize_backend(executor=PMXExecutor)

    def execute(self):
        perturbation_map: PerturbationMap = (
            self.get_workflow_object().get_perturbation_map()
        )

        # generate the B states based on the topology from pmx mutate

        for edge in perturbation_map.get_edges():
            edge_dir = os.path.join(self.work_dir, edge.get_edge_id())
            # overwrite the topology generated by pdb2gmx
            gentop_args = ["-p", "topol.top", "-o", "topol.top"]
            for branch in self.therm_cycle_branches:
                self._logger.log(
                    f"Generating B-state topology for edge {edge.get_edge_id()}, branch {branch}",
                    _LE.DEBUG,
                )
                self._backend_executor.execute(
                    command=_PE.GENTOP,
                    arguments=gentop_args,
                    check=True,
                    location=os.path.join(edge_dir, branch),
                )

            # now we have pmxtop.top in the edge dir


help_string = """
pmx gentop -h
usage: pmx [-h] [-p topol] [-o outfile] [-ff ff] [--split] [--scale_mass]
           [--scale_dih SCALE_DIH] [--norecursive]

This script fills in the B state to a topology file (itp or top) according to
the hybrid residues present in the file. If you provide a top file with
include statemets, by default the script will run through the included itp
files too; this can turned off using the --norecursive flag. You need to use
this script after having mutated a structure file with pmx mutate, and after
having passed that mutated structure through pdb2gmx.

optional arguments:
  -h, --help            show this help message and exit
  -p topol              Input topology file (itp or top). Default is
                        "topol.top"
  -o outfile            Output topology file. Default is "pmxtop.top"
  -ff ff                Force field to use. If -p is a top file, it is not
                        necessary to specify the forcefield, as it will be
                        determined automatically. If -p is an itp file, then
                        -ff is needed, and if not provided a list of available
                        ff will be shown.
  --split               Write separate topologies for the vdW and charge
                        transformations.
  --scale_mass          Scale the masses of morphing atoms so that dummies
                        have a mass of 1.
  --scale_dih SCALE_DIH
                        Scale the dihedrals that have a dummy.
  --norecursive         Whether to fill the B states also for all itp files
                        included in the provided topology file. Default is
                        True. This flag sets it to False.
"""



from shutil import copyfile
from distutils.dir_util import copy_tree
import json
import os
import sys
from typing import Any, List, Dict, Optional, Union
from copy import Error


class GenericData:
    """Container class to hold generic data of any file type"""

    def __init__(
        self,
        file_name: str,
        file_data=None,
        argument=True,
        file_id: int = None,
        extension: str = None,
    ):
        self._extension = (
            extension if extension is not None else file_name.split(".")[-1]
        )
        self._file_name = file_name
        self._file_data = file_data
        self._file_id = file_id
        # self._argument: bool = argument
        self._file_size = self.calculate_file_size()

    def get_file_name(self) -> str:
        return self._file_name

    def get_data(self) -> Any:
        return self._file_data

    def calculate_file_size(self):
        return sys.getsizeof(self._file_data)

    def get_extension(self):
        return self._extension

    def set_data(self, data):
        self._file_data = data

    def set_file_name(self, file_name):
        self._file_name = file_name

    def set_id(self, file_id):
        self._file_id = file_id

    def get_id(self):
        return self._file_id

    def set_extension(self, extension):
        self._extension = extension

    def write(self, path: str, join: bool = True, final_writeout: bool = False):
        """
        Handles all I/O operations for generic data.  Support for handling directories and symlinks
        """
        orig_path = path
        if join:
            path = os.path.join(path, self.get_file_name())

        if str(self._file_data).startswith("/"):
            # file data is a path, copy the file to the destination
            # if it's a file, its stored like this because it's large (> 2GB)
            if os.path.isfile(self._file_data):
                if not final_writeout:
                    # if this is a writeout to a step, we can simply create a simlink
                    os.symlink(self._file_data, path, target_is_directory=False)
                else:
                    # we cannot do this for the final writeout since /scratch or /tmp will eventually get cleaned
                    copyfile(self._file_data, path)

            elif os.path.isdir(self._file_data):
                # copy the entire directory to the parent dir
                copy_tree(self._file_data, orig_path)
        elif isinstance(self._file_data, list):
            with open(path, "w") as f:
                f.writelines(self._file_data)

        elif isinstance(self._file_data, str):
            with open(path, "w") as f:
                f.write(self._file_data)
        elif isinstance(self._file_data, dict):
            with open(path, "w") as f:
                f.write(json.dumps(self._file_data))
        else:
            with open(path, "wb") as f:
                f.write(self._file_data)

    def update_data(self, data):
        if sys.getsizeof(data) != self._file_size:
            self.set_data(data)

    def __repr__(self):
        return f"GenericData object - name: {self._file_name}, extension: {self._extension}."

    def __str__(self):
        return self.__repr__()


class GenericContainer:
    """Container class to hold the instances of the Generic class, separated by extension"""

    def __init__(self):
        self._file_dict: Dict[str, List] = {}

    # self._paths = []
    # self._strings = []

    def add_file(self, file: GenericData):
        ext = file.get_extension()
        file.set_id(self.get_next_file_id(ext))
        try:
            self._file_dict[ext].append(file)
        except NameError:
            self._initialise_list(ext)
            self._file_dict[ext].append(file)

    def _initialise_list(self, ext):
        self._file_dict[ext] = []

    def get_next_file_id(self, ext):
        ids = [file.get_id() for file in self.get_files_by_extension(ext)]
        if len(ids) == 0:
            return 0
        else:
            return max(ids) + 1

    def get_file_by_index(self, index):
        for file in self.get_flattened_files():
            if file.get_id() == index:
                return file

    def add_files(self, files: List[GenericData]):
        extensions = list(set([f.get_extension() for f in files]))
        if len(extensions) > 1:
            raise Error("Cannot have more than one type of file")
        else:
            if extensions[0] in self._file_dict.keys():
                self._file_dict[extensions[0]].extend(files)
            else:
                self._file_dict[extensions[0]] = files

    def get_all_files(self) -> Dict[str, List]:
        return self._file_dict

    def get_files_by_extension(self, ext: str) -> List[Optional[GenericData]]:
        if ext in self._file_dict.keys():
            return self._file_dict[ext]
        else:
            self._initialise_list(ext)
            return self._file_dict[ext]

    def get_file_names_by_extension(self, ext: str):
        try:
            return [f.get_file_name() for f in self._file_dict[ext]]
        except KeyError:
            self._initialise_list(ext)
            return [f.get_file_name() for f in self._file_dict[ext]]

    def get_file_types(self):
        return self._file_dict.keys()

    def get_flattened_files(self) -> List[GenericData]:
        rtn_files = []
        for key in self._file_dict.keys():
            for file in self._file_dict[key]:
                rtn_files.append(file)
        return rtn_files

    def get_file_by_name(self, name):
        for file in self.get_flattened_files():
            if file.get_file_name() == name:
                return file

    def clear_file_dict(self):
        self._file_dict = {}

    def get_argument_by_extension(
        self, ext, rtn_file_object=False
    ) -> Union[GenericData, str]:
        files = []
        for file in self.get_flattened_files():
            if file.get_extension() == ext:
                files.append(file)
        assert len(files) > 0, f"No files with extension {ext} were found!"
        try:
            assert len(files) == 1
        except AssertionError:
            print(
                f"Found multiple files with extension {ext}, select the index of the file to be passed as an argument\n"
            )
            print("######################")
            for idx, file in enumerate(files):
                print(f"{idx}: {file.get_file_name()}")
            print("######################")
            index = input(">>> ")
            files = [files[int(index)]]

        if not rtn_file_object:
            return files[0].get_file_name()
        else:
            return files[0]

    def write_out_all_files(self, folder):
        """flattens all files in the container and writes to the specified directory"""
        for file in self.get_flattened_files():
            file.write(folder)



import pandas as pd
import numpy as np
import random
from scipy.stats import norm
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit import DataStructs

from icolos.core.workflow_steps.active_learning.surrogate_model import SurrogateModel

from icolos.utils.enums.step_enums import StepActiveLearningEnum

_SALE = StepActiveLearningEnum()


class AcquisitionFunction:
    """class to execute all acquisition function heuristics for prospective REINVENT"""
    # TODO: assumption is that "lower scores are better" Add a variable to control this

    def __init__(
        self, surrogate: SurrogateModel, function: str, acquisition_batch_size: int
    ):
        # TODO: need to add a boolean for "highest is best"
        self.surrogate = surrogate
        self.function = function.lower()
        self.acquisition_batch_size = acquisition_batch_size

    def partition_compounds(self, original_smiles, fingerprints_pool, epsilon=0.2, beta=2, pi_epsilon=0.01):
        """partitions REINVENT's batch of SMILES into acquired and non-acquired based on the acquisition function"""
        if self.function == _SALE.RANDOM:
            df = pd.DataFrame({"smiles": original_smiles})
            shuffled = df.sample(frac=1, axis=0)
            acquired_smiles = list(shuffled["smiles"][: self.acquisition_batch_size])
            non_acquired_smiles = list(shuffled["smiles"][self.acquisition_batch_size :])

            return acquired_smiles, non_acquired_smiles

        if self.function == _SALE.GREEDY:
            """greedy"""
            fps = self._get_morgan_fingerprints(smiles=original_smiles)
            predictions = self.surrogate.predict(fps)
            df = self._get_sorted_df(smiles=original_smiles, predictions=predictions, ascending=True)
            acquired_smiles = list(df["smiles"][:self.acquisition_batch_size])
            non_acquired_smiles = list(df["smiles"][self.acquisition_batch_size:])

            return acquired_smiles, non_acquired_smiles

        if self.function == _SALE.GREEDY_EPSILON:
            """greedy-epsilon"""
            # TODO: have an "acquisition function specific parameters" block in the icolos JSON for "epsilon" parameter
            if random.random() <= epsilon:
                df = pd.DataFrame({"smiles": original_smiles})
                shuffled = df.sample(frac=1, axis=0)
                acquired_smiles = list(shuffled["smiles"][: self.acquisition_batch_size])
                non_acquired_smiles = list(shuffled["smiles"][self.acquisition_batch_size:])

                return acquired_smiles, non_acquired_smiles
            else:
                fps = self._get_morgan_fingerprints(smiles=original_smiles)
                predictions = list(self.surrogate.predict(fps))
                df = self._get_sorted_df(smiles=original_smiles, predictions=predictions, ascending=True)
                acquired_smiles = list(df["smiles"][:self.acquisition_batch_size])
                non_acquired_smiles = list(df["smiles"][self.acquisition_batch_size:])

                return acquired_smiles, non_acquired_smiles

        if self.function == _SALE.PI:
            """probability of improvement"""
            # extract the current best
            try:
                with open("current_best.txt", "r") as f:
                    current_best = float(f.readlines()[0])
            except FileNotFoundError:
                current_best = 0.0

            fps = self._get_morgan_fingerprints(smiles=original_smiles)
            predictions = self.surrogate.predict(fps)
            standard_deviations = self.surrogate.get_std(fps)

            probabilities = np.zeros(len(fps))
            for idx, (pred, std) in enumerate(zip(predictions, standard_deviations)):
                # if there is no uncertainty, then PI = 1 or 0
                # depending on whether the prediction is better than the current best
                if int(std) == 0:
                    # TODO: assumes lower value is better
                    probabilities[idx] = 1 if pred < current_best else 0

                else:
                    probabilities[idx] = norm.cdf((pred - current_best + pi_epsilon) / std)

            df = self._get_sorted_df(smiles=original_smiles, predictions=probabilities, ascending=False)
            acquired_smiles = list(df["smiles"][:self.acquisition_batch_size])
            non_acquired_smiles = list(df["smiles"][self.acquisition_batch_size:])

            return acquired_smiles, non_acquired_smiles

        if self.function == _SALE.EI:
            """expected improvement"""
            # extract the current best
            try:
                with open("current_best.txt", "r") as f:
                    current_best = float(f.readlines()[0])
            except FileNotFoundError:
                current_best = 0.0

            fps = self._get_morgan_fingerprints(smiles=original_smiles)
            predictions = self.surrogate.predict(fps)
            standard_deviations = self.surrogate.get_std(fps)

            expected_improvements = np.zeros(len(fps))
            for idx, (pred, std) in enumerate(zip(predictions, standard_deviations)):
                # if there is no uncertainty, then EI is the difference between the prediction and the current best
                if int(std) == 0:
                    # TODO: assumes lower value is better
                    expected_improvements[idx] = pred - current_best
                else:
                    numerator = pred - current_best + pi_epsilon
                    Z = numerator / std
                    expected_improvements[idx] = (numerator * norm.cdf(Z)) + (std * norm.pdf(Z))

            df = self._get_sorted_df(smiles=original_smiles, predictions=expected_improvements, ascending=True)
            acquired_smiles = list(df["smiles"][:self.acquisition_batch_size])
            non_acquired_smiles = list(df["smiles"][self.acquisition_batch_size:])

            return acquired_smiles, non_acquired_smiles

        if self.function == _SALE.TS:
            """thompson sampling"""
            fps = self._get_morgan_fingerprints(smiles=original_smiles)
            predictions = self.surrogate.predict(fps)
            standard_deviations = self.surrogate.get_std(fps)
            gaussian_predictions = np.random.default_rng().normal(predictions, standard_deviations)

            df = self._get_sorted_df(smiles=original_smiles, predictions=gaussian_predictions, ascending=True)
            acquired_smiles = list(df["smiles"][:self.acquisition_batch_size])
            non_acquired_smiles = list(df["smiles"][self.acquisition_batch_size:])

            return acquired_smiles, non_acquired_smiles

        if self.function == _SALE.UCB:
            """upper-confidence bound"""
            # TODO: have an "acquisition function specific parameters" block in the icolos JSON for "beta" parameter
            fps = self._get_morgan_fingerprints(smiles=original_smiles)
            predictions = self.surrogate.predict(fps)
            standard_deviations = self.surrogate.get_std(fps)
            ucb = predictions + beta * standard_deviations

            df = self._get_sorted_df(smiles=original_smiles, predictions=ucb, ascending=True)
            acquired_smiles = list(df["smiles"][:self.acquisition_batch_size])
            non_acquired_smiles = list(df["smiles"][self.acquisition_batch_size:])

            return acquired_smiles, non_acquired_smiles

        if self.function == _SALE.TANIMOTO:
            """Tanimoto. SMILES most dissimilar, i.e., lowest Tanimoto similarity to the pooled SMILES are acquired"""
            fps = self._get_morgan_fingerprints(smiles=original_smiles)
            similarities = np.zeros(len(fps))

            for idx, fp in enumerate(fps):
                similarities[idx] = np.mean(DataStructs.BulkTanimotoSimilarity(fp, fingerprints_pool))

            # sort by ascending Tanimoto similarities
            sorted_idx = np.argsort(similarities)
            acquired_idx = sorted_idx[:self.acquisition_batch_size]
            non_acquired_idx = sorted_idx[self.acquisition_batch_size:]

            acquired_smiles = [original_smiles[idx] for idx in acquired_idx]
            non_acquired_smiles = [original_smiles[idx] for idx in non_acquired_idx]

            return list(acquired_smiles), list(non_acquired_smiles)

        if self.function == _SALE.TANIMOTO_TRUE_BELIEVER:
            """
            Tanimoto. SMILES most dissimilar, i.e., lowest Tanimoto similarity to the pooled SMILES are acquired.
            To discourage acquiring clusters of SMILES, i.e., from the same chemical space, SMILES are acquired
            iteratively. After acquiring a SMILES, it is added to the "pool" and the same calculation is performed
            to acquire the next SMILES. This process is repeated until acquisition batch size SMILES is acquired.
            """
            from copy import deepcopy

            augmented_fps_pool = deepcopy(fingerprints_pool)

            def single_acquire(acquired_smiles, smiles_left, augmented_fps_pool):

                fps = self._get_morgan_fingerprints(smiles=smiles_left)
                similarities = np.zeros(len(fps))
                for idx, fp in enumerate(fps):
                    similarities[idx] = np.max(DataStructs.BulkTanimotoSimilarity(fp, augmented_fps_pool))

                # sort by ascending Tanimoto similarities
                sorted_idx = np.argsort(similarities)
                acquire = smiles_left[sorted_idx[0]]
                acquired_smiles.append(acquire)
                smiles_left.remove(acquire)

                augmented_fps_pool.extend(self._get_morgan_fingerprints([acquire]))

                return acquired_smiles, smiles_left, augmented_fps_pool

            acquired_smiles, smiles_left, augmented_fps_pool = single_acquire([], original_smiles, augmented_fps_pool)
            for iteration in range(self.acquisition_batch_size-1):
                acquired_smiles, smiles_left, augmented_fps_pool = single_acquire(acquired_smiles,
                                                                                  smiles_left,
                                                                                  augmented_fps_pool)

            non_acquired_smiles = smiles_left

            return list(acquired_smiles), list(non_acquired_smiles)

    def _get_morgan_fingerprints(self, smiles) -> list:
        """returns morgan fingerprints for a list of smiles"""
        molecules = [Chem.MolFromSmiles(s) for s in smiles]
        return [
            Chem.AllChem.GetMorganFingerprintAsBitVect(mol, radius=3, nBits=2048)
            for mol in molecules
        ]

    def _get_sorted_df(self, smiles, predictions, ascending):
        """
        returns a DataFrame sorted by prediction values. Can be ascending or descending,
        as specified in the method call. The sorted DataFrame will be sliced into acquired
        and non-acquired SMILES
        """
        # TODO: need "highest_is_best" to be set in class instantiation
        df = pd.DataFrame({"smiles": smiles, "predictions": predictions})

        return df.sort_values(by="predictions", ascending=ascending)



import random

import rdkit.Chem as rkc
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import SaltRemover
from rdkit.Chem import rdmolops


def _initialiseNeutralisationReactions():
    patts = (
        # Imidazoles
        ("[n+;H]", "n"),
        # Amines
        ("[N+;!H0]", "N"),
        # Carboxylic acids and alcohols
        ("[$([O-]);!$([O-][#7])]", "O"),
        # Thiols
        ("[S-;X1]", "S"),
        # Sulfonamides
        ("[$([N-;X2]S(=O)=O)]", "N"),
        # Enamines
        ("[$([N-;X2][C,N]=C)]", "N"),
        # Tetrazoles
        ("[n-]", "[nH]"),
        # Sulfoxides
        ("[$([S-]=O)]", "S"),
        # Amides
        ("[$([N-]C=O)]", "N"),
    )
    return [(Chem.MolFromSmarts(x), Chem.MolFromSmiles(y, False)) for x, y in patts]


def _neutralise_charges(mol, reactions=None):
    if reactions is None:
        reactions = _initialiseNeutralisationReactions()
    replaced = False
    for i, (reactant, product) in enumerate(reactions):
        while mol.HasSubstructMatch(reactant):
            replaced = True
            rms = AllChem.ReplaceSubstructs(mol, reactant, product)
            mol = rms[0]
    if replaced:
        return mol, True
    else:
        return mol, False


def _get_largest_fragment(mol):
    frags = rdmolops.GetMolFrags(mol, asMols=True, sanitizeFrags=True)
    maxmol = None
    for mol in frags:
        if mol is None:
            continue
        if maxmol is None:
            maxmol = mol
        if maxmol.GetNumHeavyAtoms() < mol.GetNumHeavyAtoms():
            maxmol = mol
    return maxmol


_saltremover = SaltRemover.SaltRemover()


def _valid_size(
    mol, min_heavy_atoms, max_heavy_atoms, element_list, remove_long_side_chains
):
    """Filters molecules on number of heavy atoms and atom types"""
    mol = _rare_filters(mol)
    if mol:
        correct_size = min_heavy_atoms < mol.GetNumHeavyAtoms() < max_heavy_atoms
        if not correct_size:
            return

        valid_elements = all(
            [atom.GetAtomicNum() in element_list for atom in mol.GetAtoms()]
        )
        if not valid_elements:
            return

        has_long_sidechains = False
        if remove_long_side_chains:
            # remove aliphatic side chains with at least 5 carbons not in a ring
            sma = "[CR0]-[CR0]-[CR0]-[CR0]-[CR0]"
            has_long_sidechains = mol.HasSubstructMatch(Chem.MolFromSmarts(sma))

        return correct_size and valid_elements and not has_long_sidechains


def _rare_filters(mol):
    if mol:
        ciano_filter = "[C-]#[N+]"
        oh_filter = "[OH+]"
        sulfur_filter = "[SH]"
        if (
            not mol.HasSubstructMatch(Chem.MolFromSmarts(ciano_filter))
            and not mol.HasSubstructMatch(Chem.MolFromSmarts(oh_filter))
            and not mol.HasSubstructMatch(Chem.MolFromSmarts(sulfur_filter))
        ):
            return mol


def convert_to_rdkit_smiles(smiles):
    return Chem.MolToSmiles(
        Chem.MolFromSmiles(smiles, sanitize=False), isomericSmiles=True
    )


def randomize_smiles(smiles, random_type="restricted"):
    """
    Returns a random SMILES given a SMILES of a molecule.
    :param random_type: The type (unrestricted, restricted) of randomization performed.
    :return : A random SMILES string of the same molecule or None if the molecule is invalid.
    """
    mol = Chem.MolFromSmiles(smiles)
    if not mol:
        return None

    if random_type == "unrestricted":
        return rkc.MolToSmiles(mol, canonical=False, doRandom=True, isomericSmiles=True)
    if random_type == "restricted":
        new_atom_order = list(range(mol.GetNumHeavyAtoms()))
        random.shuffle(new_atom_order)
        random_mol = rkc.RenumberAtoms(mol, newOrder=new_atom_order)
        return rkc.MolToSmiles(random_mol, canonical=False, isomericSmiles=True)
    raise ValueError("Type '{}' is not valid".format(random_type))


def to_mol(smi):
    """
    Creates a Mol object from a SMILES string.
    :param smi: SMILES string.
    :return: A Mol object or None if it's not valid.
    """
    if smi:
        return rkc.MolFromSmiles(smi)


def to_smiles(mol):
    """
    Converts a Mol object into a canonical SMILES string.
    :param mol: Mol object.
    :return: A SMILES string.
    """
    return rkc.MolToSmiles(mol, isomericSmiles=True)



import os
from abc import abstractmethod
from copy import deepcopy
from typing import Dict, List

from pydantic import BaseModel, PrivateAttr

from icolos.loggers.agentlogger import AgentLogger

from icolos.utils.enums.logging_enums import LoggingConfigEnum
from icolos.utils.enums.composite_agents_enums import WorkflowEnum

_WE = WorkflowEnum()
_LE = LoggingConfigEnum()


class AgentEnvironmentParameters(BaseModel):
    class WorkflowExportParameters(BaseModel):
        key: str
        value: str

    export: List[WorkflowExportParameters] = []


class AgentHeaderParametersSettings(BaseModel):
    remove_temporary_files: bool = True
    single_directory: bool = False


class AgentHeaderParameters(BaseModel):
    class AgentLoggingParameters(BaseModel):
        logfile: str = None

    id: str = None
    description: str = None
    logging: AgentLoggingParameters = AgentLoggingParameters()
    environment: AgentEnvironmentParameters = None
    global_variables: Dict = None
    global_settings: AgentHeaderParametersSettings = AgentHeaderParametersSettings()


class BaseAgent(BaseModel):

    # should also work without parsing the base specification here, but then IDEs will not pick up stuff below
    header: AgentHeaderParameters = AgentHeaderParameters()

    class Config:
        underscore_attrs_are_private = True

    _logger = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._logger = AgentLogger()

    def _export_env_variables(self):
        for var in self.header.environment.export:
            key = str(var.key)
            value = os.path.expandvars(str(var.value))
            os.environ[key] = value
            self._logger.log(f"Exported variable {key} with value {value}.", _LE.DEBUG)

    def initialize(self):
        self._export_env_variables()

    def _nested_update(self, inp, pattern: str, replacement: str):
        if isinstance(inp, dict):
            items = inp.items()
        elif isinstance(inp, (list, tuple)):
            items = enumerate(inp)
        elif isinstance(inp, str):
            return inp.replace(pattern, replacement)
        else:
            return inp

        for key, value in items:
            inp[key] = self._nested_update(value, pattern, replacement)
        return inp

    def _update_global_variables(self, conf: dict) -> dict:
        conf = deepcopy(conf)
        if self.header.global_variables is not None:
            for key, value in self.header.global_variables.items():
                pattern = "{" + key + "}"
                self._nested_update(inp=conf, pattern=pattern, replacement=value)
                self._logger.log(
                    f"Updated global variable {key} with value {value}.", _LE.DEBUG
                )
        return conf

    @abstractmethod
    def execute(self):
        raise NotImplementedError

    def is_valid(self) -> bool:
        raise NotImplementedError

    def set_id(self, id: str):
        self.header.id = id

    def get_id(self) -> str:
        return self.header.id

    def set_description(self, description: str):
        self.header.description = description

    def get_description(self) -> str:
        return self.header.description
